# ***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "CompoundTools._CommandExplodeCompound"
__author__ = "DeepSOIC"
__url__ = "http://www.freecadweb.org"
__doc__ = "ExplodeCompound: create a bunch of CompoundFilter objects to split a compound into pieces."

from .Explode import explodeCompound

import FreeCAD
if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtGui
    from PySide import QtCore


# translation-related code
    try:
        _fromUtf8 = QtCore.QString.fromUtf8
    except Exception:
        def _fromUtf8(s):
            return s
    try:
        _encoding = QtGui.QApplication.UnicodeUTF8
        def _translate(context, text, disambig):
            return QtGui.QApplication.translate(context, text, disambig, _encoding)
    except AttributeError:
        def _translate(context, text, disambig):
            return QtGui.QApplication.translate(context, text, disambig)


# command class
class _CommandExplodeCompound:
    "Command to explode a compound"
    def GetResources(self):
        return {'Pixmap': "Part_ExplodeCompound",
                'MenuText': QtCore.QT_TRANSLATE_NOOP("Part_ExplodeCompound", "Explode compound"),
                'Accel': "",
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("Part_ExplodeCompound",
                                                    "Split up a compound of shapes into separate objects.\n"
                                                    "It will create a 'Compound Filter' for each shape.")}

    def Activated(self):
        if len(FreeCADGui.Selection.getSelection()) == 1:
            cmdExplode()
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(_translate("Part_ExplodeCompound", "First select a shape that is a compound.", None))
            mb.setWindowTitle(_translate("Part_ExplodeCompound", "Bad selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


if FreeCAD.GuiUp:
    FreeCADGui.addCommand('Part_ExplodeCompound', _CommandExplodeCompound())




def cmdExplode():
    FreeCAD.ActiveDocument.openTransaction("Explode")
    try:
        sel = FreeCADGui.Selection.getSelectionEx()
        if len(sel) != 1:
            raise ValueError("Bad selection","More than one object selected. You have selected {num} objects.".format(num= len(sel)))
        obj = sel[0].Object
        FreeCADGui.addModule("CompoundTools.Explode")
        FreeCADGui.doCommand("input_obj = App.ActiveDocument."+obj.Name)
        FreeCADGui.doCommand("CompoundTools.Explode.explodeCompound(input_obj)")
        FreeCADGui.doCommand("input_obj.ViewObject.hide()")
    except Exception:
        FreeCAD.ActiveDocument.abortTransaction()
        raise
        
    FreeCAD.ActiveDocument.commitTransaction()
    FreeCADGui.doCommand("App.ActiveDocument.recompute()")
    



# -*- coding: utf-8 -*-
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from __future__ import print_function

import FreeCAD
import Path
import PathScripts.PathCircularHoleBase as PathCircularHoleBase
import PathScripts.PathGeom as PathGeom
import PathScripts.PathLog as PathLog
import PathScripts.PathOp as PathOp
import PathScripts.PathUtils as PathUtils
import math

from PySide import QtCore

__title__ = "Path Thread Milling Operation"
__author__ = "sliptonic (Brad Collette)"
__url__ = "http://www.freecadweb.org"
__doc__ = "Path thread milling operation."

PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())
#PathLog.trackModule(PathLog.thisModule())

# Qt translation handling
def translate(context, text, disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)


def radiiInternal(majorDia, minorDia, toolDia, toolCrest = None):
    '''internlThreadRadius(majorDia, minorDia, toolDia, toolCrest) ... returns the maximum radius for thread.'''
    PathLog.track(majorDia, minorDia, toolDia, toolCrest)
    if toolCrest is None:
        toolCrest = 0.0
    # As it turns out metric and imperial standard threads follow the same rules.
    # The determining factor is the height of the full 60 degree triangle H.
    # - The minor diameter is 1/4 * H smaller than the pitch diameter.
    # - The major diameter is 3/8 * H bigger than the pitch diameter
    # Since we already have the outer diameter it's simpler to just add 1/8 * H
    # to get the outer tip of the thread.
    H = ((majorDia - minorDia) / 2.0 ) * 1.6             # (D - d)/2 = 5/8 * H
    outerTip = majorDia / 2.0 + H / 8.0
    # Compensate for the crest of the tool
    toolTip = outerTip - toolCrest * 0.8660254037844386  # math.sqrt(3)/2 ... 60deg triangle height
    return ((minorDia - toolDia) / 2.0, toolTip - toolDia / 2.0)

def threadPasses(count, radii, majorDia, minorDia, toolDia, toolCrest = None):
    PathLog.track(count, radii, majorDia, minorDia, toolDia, toolCrest)
    minor, major = radii(majorDia, minorDia, toolDia, toolCrest)
    dr  = float(major - minor) / count
    return [major - dr * (count - (i + 1)) for i in range(count)]

class _InternalThread(object):
    '''Helper class for dealing with different thread types'''
    def __init__(self, cmd, zStart, zFinal, pitch):
        self.cmd = cmd
        if zStart < zFinal:
            self.pitch = pitch
        else:
            self.pitch = -pitch
        self.hPitch = self.pitch / 2
        self.zStart = zStart
        self.zFinal = zFinal

    def overshoots(self, z):
        '''overshoots(z) ... returns true if adding another half helix goes beyond the thread bounds'''
        if self.pitch < 0:
            return z + self.hPitch < self.zFinal
        return z + self.hPitch > self.zFinal

    def adjustX(self, x, dx):
        '''adjustX(x, dx) ... move x by dx, the direction depends on the thread settings'''
        if self.isG3() == (self.pitch > 0):
            return x + dx
        return x - dx

    def adjustY(self, y, dy):
        '''adjustY(y, dy) ... move y by dy, the direction depends on the thread settings'''
        if self.isG3():
            return y - dy
        return y - dy

    def isG3(self):
        '''isG3() ... returns True if this is a G3 command'''
        return self.cmd in ['G3', 'G03', 'g3', 'g03']

    def isUp(self):
        '''isUp() ... returns True if the thread goes from the bottom up'''
        return self.pitch > 0

def internalThreadCommands(loc, cmd, zStart, zFinal, pitch, radius, leadInOut):
    '''internalThreadCommands(loc, cmd, zStart, zFinal, pitch, radius) ... returns the g-code to mill the given internal thread'''
    thread = _InternalThread(cmd, zStart, zFinal, pitch)

    yMin = loc.y - radius
    yMax = loc.y + radius

    path = []
    # at this point the tool is at a safe height (depending on the previous thread), so we can move
    # into position first, and then drop to the start height. If there is any material in the way this
    # op hasn't been setup properly.
    path.append(Path.Command('G0', {'X': loc.x, 'Y': loc.y}))
    path.append(Path.Command('G0', {'Z': thread.zStart}))
    if leadInOut:
        path.append(Path.Command(thread.cmd, {'Y': yMax, 'J': (yMax - loc.y) / 2}))
    else:
        path.append(Path.Command('G1', {'Y': yMax}))

    z = thread.zStart
    r = -radius
    i = 0
    while True:
        z = thread.zStart + i*thread.hPitch
        if thread.overshoots(z):
            break
        if 0 == (i & 0x01):
            y = yMin
        else:
            y = yMax
        path.append(Path.Command(thread.cmd, {'Y': y, 'Z': z + thread.hPitch, 'J': r}))
        r = -r
        i = i + 1


    z = thread.zStart + i*thread.hPitch
    if PathGeom.isRoughly(z, thread.zFinal):
        x = loc.x
    else:
        n = math.fabs(thread.zFinal - thread.zStart) / thread.hPitch
        k = n - int(n)
        dy = math.cos(k * math.pi)
        dx = math.sin(k * math.pi)
        y = thread.adjustY(loc.y, r * dy)
        x = thread.adjustX(loc.x, r * dx)
        path.append(Path.Command(thread.cmd, {'X': x, 'Y': y, 'Z': thread.zFinal, 'J': r}))

    if leadInOut:
        path.append(Path.Command(thread.cmd, {'X': loc.x, 'Y': loc.y, 'I': (loc.x - x) / 2, 'J': (loc.y - y) / 2})) 
    else:
        path.append(Path.Command('G1', {'X': loc.x, 'Y': loc.y}))
    return path

class ObjectThreadMilling(PathCircularHoleBase.ObjectOp):
    '''Proxy object for thread milling operation.'''

    LeftHand                    = 'LeftHand'
    RightHand                   = 'RightHand'
    ThreadTypeCustom            = 'Custom'
    ThreadTypeMetricInternal    = 'Metric - internal'
    ThreadTypeImperialInternal  = 'Imperial - internal'
    DirectionClimb              = 'Climb'
    DirectionConventional       = 'Conventional'

    ThreadOrientations = [LeftHand, RightHand]
    ThreadTypes = [ThreadTypeCustom, ThreadTypeMetricInternal, ThreadTypeImperialInternal]
    Directions =  [DirectionClimb, DirectionConventional]

    def circularHoleFeatures(self, obj):
        return PathOp.FeatureBaseGeometry

    def initCircularHoleOperation(self, obj):
        obj.addProperty("App::PropertyEnumeration", "ThreadOrientation", "Thread", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Set thread orientation"))
        obj.ThreadOrientation = self.ThreadOrientations
        obj.addProperty("App::PropertyEnumeration", "ThreadType", "Thread", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Currently only internal"))
        obj.ThreadType = self.ThreadTypes
        obj.addProperty("App::PropertyString", "ThreadName", "Thread", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Defines which standard thread was chosen"))
        obj.addProperty("App::PropertyLength", "MajorDiameter", "Thread", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Set thread's major diameter"))
        obj.addProperty("App::PropertyLength", "MinorDiameter", "Thread", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Set thread's minor diameter"))
        obj.addProperty("App::PropertyLength", "Pitch", "Thread", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Set thread's pitch - used for metric threads"))
        obj.addProperty("App::PropertyInteger", "TPI", "Thread", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Set thread's TPI (turns per inch) - used for imperial threads"))
        obj.addProperty("App::PropertyInteger", "ThreadFit", "Thread", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Set how many passes are used to cut the thread"))
        obj.addProperty("App::PropertyInteger", "Passes", "Operation", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Set how many passes are used to cut the thread"))
        obj.addProperty("App::PropertyEnumeration", "Direction", "Operation", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Direction of thread cutting operation"))
        obj.addProperty("App::PropertyBool", "LeadInOut", "Operation", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Set to True to get lead in and lead out arcs at the start and end of the thread cut"))
        obj.addProperty("App::PropertyLink", "ClearanceOp", "Operation", QtCore.QT_TRANSLATE_NOOP("PathThreadMilling", "Operation to clear the inside of the thread"))
        obj.Direction = self.Directions

        # Rotation related properties
        if not hasattr(obj, 'EnableRotation'):
            obj.addProperty("App::PropertyEnumeration", "EnableRotation", "Rotation", QtCore.QT_TRANSLATE_NOOP("App::Property", "Enable rotation to gain access to pockets/areas not normal to Z axis."))
            obj.EnableRotation = ['Off', 'A(x)', 'B(y)', 'A & B']


    def threadStartDepth(self, obj):
        if obj.ThreadOrientation == self.RightHand:
            if obj.Direction == self.DirectionClimb:
                PathLog.track(obj.Label, obj.FinalDepth)
                return obj.FinalDepth
            PathLog.track(obj.Label, obj.StartDepth)
            return obj.StartDepth
        if obj.Direction == self.DirectionClimb:
            PathLog.track(obj.Label, obj.StartDepth)
            return obj.StartDepth
        PathLog.track(obj.Label, obj.FinalDepth)
        return obj.FinalDepth

    def threadFinalDepth(self, obj):
        PathLog.track(obj.Label)
        if obj.ThreadOrientation == self.RightHand:
            if obj.Direction == self.DirectionClimb:
                PathLog.track(obj.Label, obj.StartDepth)
                return obj.StartDepth
            PathLog.track(obj.Label, obj.FinalDepth)
            return obj.FinalDepth
        if obj.Direction == self.DirectionClimb:
            PathLog.track(obj.Label, obj.FinalDepth)
            return obj.FinalDepth
        PathLog.track(obj.Label, obj.StartDepth)
        return obj.StartDepth

    def threadDirectionCmd(self, obj):
        PathLog.track(obj.Label)
        if obj.ThreadOrientation == self.RightHand:
            if obj.Direction == self.DirectionClimb:
                PathLog.track(obj.Label, 'G2')
                return 'G2'
            PathLog.track(obj.Label, 'G3')
            return 'G3'
        if obj.Direction == self.DirectionClimb:
            PathLog.track(obj.Label, 'G3')
            return 'G3'
        PathLog.track(obj.Label, 'G2')
        return 'G2'

    def threadSetup(self, obj):
        # the thing to remember is that Climb, for an internal thread must always be G3
        if obj.Direction == self.DirectionClimb:
            if obj.ThreadOrientation == self.RightHand:
                return ('G3', obj.FinalDepth.Value, obj.StartDepth.Value)
            return ('G3', obj.StartDepth.Value, obj.FinalDepth.Value)
        if obj.ThreadOrientation == self.RightHand:
            return ('G2', obj.StartDepth.Value, obj.FinalDepth.Value)
        return ('G2', obj.FinalDepth.Value, obj.StartDepth.Value)

    def threadPassRadii(self, obj):
        PathLog.track(obj.Label)
        rMajor = (obj.MajorDiameter.Value - self.tool.Diameter) / 2.0
        rMinor = (obj.MinorDiameter.Value - self.tool.Diameter) / 2.0
        if obj.Passes < 1:
            obj.Passes = 1
        rPass  = (rMajor - rMinor) / obj.Passes
        passes = [rMajor]
        for i in range(1, obj.Passes):
            passes.append(rMajor - rPass * i)
        return list(reversed(passes))

    def executeThreadMill(self, obj, loc, gcode, zStart, zFinal, pitch):
        PathLog.track(obj.Label, loc, gcode, zStart, zFinal, pitch)

        self.commandlist.append(Path.Command('G0', {'Z': obj.ClearanceHeight.Value, 'F': self.vertRapid}))

        for radius in threadPasses(obj.Passes, radiiInternal, obj.MajorDiameter.Value, obj.MinorDiameter.Value, float(self.tool.Diameter), float(self.tool.Crest)):
            commands = internalThreadCommands(loc, gcode, zStart, zFinal, pitch, radius, obj.LeadInOut)
            for cmd in commands:
                p = cmd.Parameters
                if cmd.Name in ['G0']:
                    p.update({'F': self.vertRapid})
                if cmd.Name in ['G1', 'G2', 'G3']:
                    p.update({'F': self.horizFeed})
                cmd.Parameters = p
            self.commandlist.extend(commands)

        self.commandlist.append(Path.Command('G0', {'Z': obj.ClearanceHeight.Value, 'F': self.vertRapid}))

    def circularHoleExecute(self, obj, holes):
        PathLog.track()
        if self.isToolSupported(obj, self.tool):
            self.commandlist.append(Path.Command("(Begin Thread Milling)"))

            (cmd, zStart, zFinal) = self.threadSetup(obj)
            pitch = obj.Pitch.Value
            if obj.TPI > 0:
                pitch = 25.4 / obj.TPI
            if pitch <= 0:
                PathLog.error("Cannot create thread with pitch {}".format(pitch))
                return

            # rapid to clearance height
            for loc in holes:
                self.executeThreadMill(obj, FreeCAD.Vector(loc['x'], loc['y'], 0), cmd, zStart, zFinal, pitch)
        else:
            PathLog.error("No suitable Tool found for thread milling operation")


    def opSetDefaultValues(self, obj, job):
        obj.ThreadOrientation = self.RightHand
        obj.ThreadType = self.ThreadTypeMetricInternal
        obj.ThreadFit = 50
        obj.Pitch = 1
        obj.TPI = 0
        obj.Passes = 1
        obj.Direction = self.DirectionClimb
        obj.LeadInOut = True

    def isToolSupported(self, obj, tool):
        '''Thread milling only supports thread milling cutters.'''
        return hasattr(tool, 'Diameter') and hasattr(tool, 'Crest')


def SetupProperties():
    setup = []
    setup.append("ThreadOrientation")
    setup.append("ThreadType")
    setup.append("ThreadName")
    setup.append("ThreadFit")
    setup.append("MajorDiameter")
    setup.append("MinorDiameter")
    setup.append("Pitch")
    setup.append("TPI")
    setup.append("Passes")
    setup.append("Direction")
    setup.append("LeadInOut")
    return setup


def Create(name, obj=None):
    '''Create(name) ... Creates and returns a thread milling operation.'''
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectThreadMilling(obj, name)
    if obj.Proxy:
        obj.Proxy.findAllHoles(obj)
    return obj




# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Layer objects."""
## @package make_layer
# \ingroup draftmake
# \brief Provides functions to create Layer objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils

from draftutils.messages import _msg, _err
from draftutils.translate import translate
from draftobjects.layer import (Layer,
                                LayerContainer)

if App.GuiUp:
    from draftviewproviders.view_layer import (ViewProviderLayer,
                                               ViewProviderLayerContainer)

view_group = App.ParamGet("User parameter:BaseApp/Preferences/View")


def get_layer_container():
    """Return a group object to put layers in.

    Returns
    -------
    App::DocumentObjectGroupPython
        The existing group object named `'LayerContainer'`
        of type `LayerContainer`.
        If it doesn't exist it will create it with this default Name.
    """
    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft","No active document. Aborting."))
        return None

    for obj in doc.Objects:
        if obj.Name == "LayerContainer":
            return obj

    new_obj = doc.addObject("App::DocumentObjectGroupPython",
                            "LayerContainer")
    new_obj.Label = translate("draft", "Layers")

    LayerContainer(new_obj)

    if App.GuiUp:
        ViewProviderLayerContainer(new_obj.ViewObject)

    return new_obj


def getLayerContainer():
    """Get the Layer container. DEPRECATED. Use 'get_layer_container'."""
    utils.use_instead("get_layer_container")

    return get_layer_container()


def make_layer(name=None,
               line_color=None, shape_color=None,
               line_width=2.0,
               draw_style="Solid", transparency=0):
    """Create a Layer object in the active document.

    If a layer container named `'LayerContainer'` does not exist,
    it is created with this name.

    A layer controls the view properties of the objects inside the layer,
    so all parameters except for `name` only apply if the graphical interface
    is up.

    Parameters
    ----------
    name: str, optional
        It is used to set the layer's `Label` (user editable).
        It defaults to `None`, in which case the `Label`
        is set to `'Layer'` or to its translation in the current language.

    line_color: tuple, optional
        It defaults to `None`, in which case it uses the value of the parameter
        `User parameter:BaseApp/Preferences/View/DefaultShapeLineColor`.
        If it is given, it should be a tuple of three
        floating point values from 0.0 to 1.0.

    shape_color: tuple, optional
        It defaults to `None`, in which case it uses the value of the parameter
        `User parameter:BaseApp/Preferences/View/DefaultShapeColor`.
        If it is given, it should be a tuple of three
        floating point values from 0.0 to 1.0.

    line_width: float, optional
        It defaults to 2.0.
        It determines the width of the edges of the objects contained
        in the layer.

    draw_style: str, optional
        It defaults to `'Solid'`.
        It determines the style of the edges of the objects contained
        in the layer.
        If it is given, it should be 'Solid', 'Dashed', 'Dotted',
        or 'Dashdot'.

    transparency: int, optional
        It defaults to 0.
        It should be an integer value from 0 (completely opaque)
        to 100 (completely transparent).

    Return
    ------
    App::FeaturePython
        A scripted object of type `'Layer'`.
        This object does not have a `Shape` attribute.
        Modifying the view properties of this object will affect the objects
        inside of it.

    None
        If there is a problem it will return `None`.
    """
    _name = "make_layer"
    utils.print_header(_name, translate("draft","Layer"))

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft","No active document. Aborting."))
        return None

    if name:
        _msg("name: {}".format(name))
        try:
            utils.type_check([(name, str)], name=_name)
        except TypeError:
            _err(translate("draft","Wrong input: it must be a string."))
            return None
    else:
        name = translate("draft", "Layer")

    if line_color:
        _msg("line_color: {}".format(line_color))
        try:
            utils.type_check([(line_color, tuple)], name=_name)
        except TypeError:
            _err(translate("draft","Wrong input: must be a tuple of three floats 0.0 to 1.0."))
            return None

        if not all(isinstance(color, (int, float)) for color in line_color):
            _err(translate("draft","Wrong input: must be a tuple of three floats 0.0 to 1.0."))
            return None
    else:
        c = view_group.GetUnsigned("DefaultShapeLineColor", 255)
        line_color = (((c >> 24) & 0xFF) / 255,
                      ((c >> 16) & 0xFF) / 255,
                      ((c >> 8) & 0xFF) / 255)

    if shape_color:
        _msg("shape_color: {}".format(shape_color))
        try:
            utils.type_check([(shape_color, tuple)], name=_name)
        except TypeError:
            _err(translate("draft","Wrong input: must be a tuple of three floats 0.0 to 1.0."))
            return None

        if not all(isinstance(color, (int, float)) for color in shape_color):
            _err(translate("draft","Wrong input: must be a tuple of three floats 0.0 to 1.0."))
            return None
    else:
        c = view_group.GetUnsigned("DefaultShapeColor", 4294967295)
        shape_color = (((c >> 24) & 0xFF) / 255,
                       ((c >> 16) & 0xFF) / 255,
                       ((c >> 8) & 0xFF) / 255)

    _msg("line_width: {}".format(line_width))
    try:
        utils.type_check([(line_width, (int, float))], name=_name)
        line_width = float(abs(line_width))
    except TypeError:
        _err(translate("draft","Wrong input: must be a number."))
        return None

    _msg("draw_style: {}".format(draw_style))
    try:
        utils.type_check([(draw_style, str)], name=_name)
    except TypeError:
        _err(translate("draft","Wrong input: must be 'Solid', 'Dashed', 'Dotted', or 'Dashdot'."))
        return None

    if draw_style not in ('Solid', 'Dashed', 'Dotted', 'Dashdot'):
        _err(translate("draft","Wrong input: must be 'Solid', 'Dashed', 'Dotted', or 'Dashdot'."))
        return None

    _msg("transparency: {}".format(transparency))
    try:
        utils.type_check([(transparency, (int, float))], name=_name)
        transparency = int(abs(transparency))
    except TypeError:
        _err(translate("draft","Wrong input: must be a number between 0 and 100."))
        return None

    new_obj = doc.addObject("App::FeaturePython", "Layer")
    Layer(new_obj)

    new_obj.Label = name

    if App.GuiUp:
        ViewProviderLayer(new_obj.ViewObject)

        new_obj.ViewObject.LineColor = line_color
        new_obj.ViewObject.ShapeColor = shape_color
        new_obj.ViewObject.LineWidth = line_width
        new_obj.ViewObject.DrawStyle = draw_style
        new_obj.ViewObject.Transparency = transparency

    container = get_layer_container()
    container.addObject(new_obj)

    return new_obj


def makeLayer(name=None, linecolor=None, drawstyle=None,
              shapecolor=None, transparency=None):
    """Create a Layer. DEPRECATED. Use 'make_layer'."""
    utils.use_instead("make_layer")

    if not drawstyle:
        drawstyle = "Solid"

    if not transparency:
        transparency = 0

    return make_layer(name,
                      line_color=linecolor, shape_color=shapecolor,
                      draw_style=drawstyle, transparency=transparency)

## @}



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the Label object."""
## @package view_label
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the Label object.

## \addtogroup draftviewproviders
# @{
import math
import sys
import pivy.coin as coin
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import DraftVecUtils
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils

from draftviewproviders.view_draft_annotation \
    import ViewProviderDraftAnnotation

if App.GuiUp:
    import FreeCADGui as Gui

param = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")


class ViewProviderLabel(ViewProviderDraftAnnotation):
    """Viewprovider for the Label annotation object."""

    def __init__(self, vobj):
        super(ViewProviderLabel, self).__init__(vobj)

        self.set_properties(vobj)
        self.Object = vobj.Object
        vobj.Proxy = self

    def set_properties(self, vobj):
        """Set the properties only if they don't already exist."""
        super(ViewProviderLabel, self).set_properties(vobj)

        properties = vobj.PropertiesList
        self.set_text_properties(vobj, properties)
        self.set_graphics_properties(vobj, properties)

    def set_text_properties(self, vobj, properties):
        """Set text properties only if they don't already exist."""
        if "TextSize" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The size of the text")
            vobj.addProperty("App::PropertyLength",
                             "TextSize",
                             "Text",
                             _tip)
            vobj.TextSize = utils.get_param("textheight", 1)

        if "TextFont" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The font of the text")
            vobj.addProperty("App::PropertyFont",
                             "TextFont",
                             "Text",
                             _tip)
            vobj.TextFont = utils.get_param("textfont")

        if "TextAlignment" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The vertical alignment of the text")
            vobj.addProperty("App::PropertyEnumeration",
                             "TextAlignment",
                             "Text",
                             _tip)
            vobj.TextAlignment = ["Top", "Middle", "Bottom"]
            vobj.TextAlignment = "Bottom"

        if "TextColor" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "Text color")
            vobj.addProperty("App::PropertyColor",
                             "TextColor",
                             "Text",
                             _tip)

        if "MaxChars" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The maximum number of characters "
                                     "on each line of the text box")
            vobj.addProperty("App::PropertyInteger",
                             "MaxChars",
                             "Text",
                             _tip)

        if "Justification" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The vertical alignment of the text")
            vobj.addProperty("App::PropertyEnumeration",
                             "Justification",
                             "Text",
                             _tip)
            vobj.Justification = ["Left", "Center", "Right"]

        if "LineSpacing" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "Line spacing (relative to font size)")
            vobj.addProperty("App::PropertyFloat",
                             "LineSpacing",
                             "Text",
                             _tip)
            vobj.LineSpacing = 1.0

    def set_graphics_properties(self, vobj, properties):
        """Set graphics properties only if they don't already exist."""
        if "ArrowSize" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The size of the arrow")
            vobj.addProperty("App::PropertyLength",
                             "ArrowSize",
                             "Graphics",
                             _tip)
            vobj.ArrowSize = utils.get_param("arrowsize", 1)

        if "ArrowType" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The type of arrow of this label")
            vobj.addProperty("App::PropertyEnumeration",
                             "ArrowType",
                             "Graphics",
                             _tip)
            vobj.ArrowType = utils.ARROW_TYPES
            vobj.ArrowType = utils.ARROW_TYPES[utils.get_param("dimsymbol")]

        if "Frame" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The type of frame around the text "
                                     "of this object")
            vobj.addProperty("App::PropertyEnumeration",
                             "Frame",
                             "Graphics",
                             _tip)
            vobj.Frame = ["None", "Rectangle"]

        if "Line" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "Display a leader line or not")
            vobj.addProperty("App::PropertyBool",
                             "Line",
                             "Graphics",
                             _tip)
            vobj.Line = True

        if "LineWidth" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "Line width")
            vobj.addProperty("App::PropertyFloat",
                             "LineWidth",
                             "Graphics",
                             _tip)
            vobj.LineWidth = utils.get_param("linewidth", 1)

        if "LineColor" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "Line color")
            vobj.addProperty("App::PropertyColor",
                             "LineColor",
                             "Graphics",
                             _tip)

    def getIcon(self):
        """Return the path to the icon used by the viewprovider."""
        return ":/icons/Draft_Label.svg"

    def claimChildren(self):
        """Return objects that will be placed under it in the tree view."""
        return []

    def attach(self, vobj):
        """Set up the scene sub-graph of the viewprovider."""
        # Attributes of the Coin scenegraph
        self.arrow = coin.SoSeparator()
        self.arrowpos = coin.SoTransform()
        self.arrow.addChild(self.arrowpos)

        self.matline = coin.SoMaterial()
        self.drawstyle = coin.SoDrawStyle()
        self.drawstyle.style = coin.SoDrawStyle.LINES

        self.lcoords = coin.SoCoordinate3()
        self.line = coin.SoType.fromName("SoBrepEdgeSet").createInstance()

        self.mattext = coin.SoMaterial()
        self.textpos = coin.SoTransform()
        self.font = coin.SoFont()
        self.text2d = coin.SoText2()  # Faces the camera always
        self.text3d = coin.SoAsciiText()  # Can be oriented in 3D space

        self.fcoords = coin.SoCoordinate3()
        self.frame = coin.SoType.fromName("SoBrepEdgeSet").createInstance()
        self.lineswitch = coin.SoSwitch()

        self.symbol = gui_utils.dim_symbol()

        textdrawstyle = coin.SoDrawStyle()
        textdrawstyle.style = coin.SoDrawStyle.FILLED

        # The text string needs to be initialized to something,
        # otherwise it crashes
        self.text2d.string = self.text3d.string = "Label"
        self.text2d.justification = coin.SoText2.RIGHT
        self.text3d.justification = coin.SoAsciiText.RIGHT

        switchnode = coin.SoSeparator()
        switchnode.addChild(self.line)
        switchnode.addChild(self.arrow)
        self.lineswitch.addChild(switchnode)
        self.lineswitch.whichChild = 0

        self.node2d = coin.SoGroup()
        self.node2d.addChild(self.matline)
        self.node2d.addChild(self.arrow)
        self.node2d.addChild(self.drawstyle)
        self.node2d.addChild(self.lcoords)
        self.node2d.addChild(self.lineswitch)
        self.node2d.addChild(self.mattext)
        self.node2d.addChild(textdrawstyle)
        self.node2d.addChild(self.textpos)
        self.node2d.addChild(self.font)
        self.node2d.addChild(self.text2d)
        self.node2d.addChild(self.fcoords)
        self.node2d.addChild(self.frame)

        self.node3d = coin.SoGroup()
        self.node3d.addChild(self.matline)
        self.node3d.addChild(self.arrow)
        self.node3d.addChild(self.drawstyle)
        self.node3d.addChild(self.lcoords)
        self.node3d.addChild(self.lineswitch)
        self.node3d.addChild(self.mattext)
        self.node3d.addChild(textdrawstyle)
        self.node3d.addChild(self.textpos)
        self.node3d.addChild(self.font)
        self.node3d.addChild(self.text3d)
        self.node3d.addChild(self.fcoords)
        self.node3d.addChild(self.frame)

        vobj.addDisplayMode(self.node2d, "2D text")
        vobj.addDisplayMode(self.node3d, "3D text")
        self.onChanged(vobj, "LineColor")
        self.onChanged(vobj, "TextColor")
        self.onChanged(vobj, "ArrowSize")
        self.onChanged(vobj, "Line")
        # self.onChanged(vobj, "ScaleMultiplier")

    def getDisplayModes(self, vobj):
        """Return the display modes that this viewprovider supports."""
        return ["2D text", "3D text"]

    def getDefaultDisplayMode(self):
        """Return the default display mode."""
        return "3D text"

    def setDisplayMode(self, mode):
        """Return the saved display mode."""
        return mode

    def updateData(self, obj, prop):
        """Execute when a property from the Proxy class is changed."""
        if prop == "Points":
            n_points = len(obj.Points)
            if n_points >= 2:
                self.line.coordIndex.deleteValues(0)
                self.lcoords.point.setValues(obj.Points)
                self.line.coordIndex.setValues(0,
                                               n_points,
                                               range(n_points))
                self.onChanged(obj.ViewObject, "TextSize")
                self.onChanged(obj.ViewObject, "ArrowType")

            if obj.StraightDistance > 0:
                self.text2d.justification = coin.SoText2.RIGHT
                self.text3d.justification = coin.SoAsciiText.RIGHT
            else:
                self.text2d.justification = coin.SoText2.LEFT
                self.text3d.justification = coin.SoAsciiText.LEFT

            self.onChanged(obj.ViewObject, "TextAlignment")
            self.onChanged(obj.ViewObject, "Frame")

        elif prop == "Text" and obj.Text:
            self.text2d.string.setValue("")
            self.text3d.string.setValue("")

            if sys.version_info.major >= 3:
                _list = [l for l in obj.Text if l]
            else:
                _list = [l.encode("utf8") for l in obj.Text if l]

            self.text2d.string.setValues(_list)
            self.text3d.string.setValues(_list)
            self.onChanged(obj.ViewObject, "TextAlignment")
            self.onChanged(obj.ViewObject, "Frame")

    def onChanged(self, vobj, prop):
        """Execute when a view property is changed."""
        super(ViewProviderLabel, self).onChanged(vobj, prop)

        obj = vobj.Object
        properties = vobj.PropertiesList

        if prop == "ScaleMultiplier" and "ScaleMultiplier" in properties:
            if "TextSize" in properties and "TextAlignment" in properties:
                self.update_label(obj, vobj)
            if "ArrowSize" in properties:
                s = vobj.ArrowSize.Value * vobj.ScaleMultiplier
                if s:
                    self.arrowpos.scaleFactor.setValue((s, s, s))

        elif prop == "LineColor" and "LineColor" in properties:
            col = vobj.LineColor
            self.matline.diffuseColor.setValue([col[0], col[1], col[2]])

        elif prop == "TextColor" and "TextColor" in properties:
            col = vobj.TextColor
            self.mattext.diffuseColor.setValue([col[0], col[1], col[2]])

        elif prop == "LineWidth" and "LineWidth" in properties:
            self.drawstyle.lineWidth = vobj.LineWidth

        elif prop == "TextFont" and "TextFont" in properties:
            self.font.name = vobj.TextFont.encode("utf8")

        elif (prop in ["TextSize", "TextAlignment"]
              and "ScaleMultiplier" in properties
              and "TextSize" in properties
              and "TextAlignment" in properties):
            self.update_label(obj, vobj)

        elif prop == "Line" and "Line" in properties:
            if vobj.Line:
                self.lineswitch.whichChild = 0
            else:
                self.lineswitch.whichChild = -1

        elif prop == "ArrowType" and "ArrowType" in properties:
            if len(obj.Points) > 1:
                self.update_arrow(obj, vobj)

        elif (prop == "ArrowSize" and "ArrowSize" in properties
              and "ScaleMultiplier" in properties):
            s = vobj.ArrowSize.Value * vobj.ScaleMultiplier
            if s:
                self.arrowpos.scaleFactor.setValue((s, s, s))

        elif prop == "Frame" and "Frame" in properties:
            self.frame.coordIndex.deleteValues(0)

            if vobj.Frame == "Rectangle":
                self.draw_frame(obj, vobj)

        elif prop in "Justification" and "Justification" in properties:
            if vobj.Justification == "Left":
                self.text2d.justification = coin.SoText2.LEFT
                self.text3d.justification = coin.SoAsciiText.LEFT
            elif vobj.Justification == "Right":
                self.text2d.justification = coin.SoText2.RIGHT
                self.text3d.justification = coin.SoAsciiText.RIGHT
            else:
                self.text2d.justification = coin.SoText2.CENTER
                self.text3d.justification = coin.SoAsciiText.CENTER

        elif prop == "LineSpacing" and "LineSpacing" in properties:
            self.text2d.spacing = vobj.LineSpacing
            self.text3d.spacing = vobj.LineSpacing

    def get_text_size(self, vobj):
        """Return the bunding box of the text element."""
        if vobj.DisplayMode == "3D text":
            text = self.text3d
        else:
            text = self.text2d

        view = Gui.ActiveDocument.ActiveView
        region = view.getViewer().getSoRenderManager().getViewportRegion()
        action = coin.SoGetBoundingBoxAction(region)
        text.getBoundingBox(action)

        return action.getBoundingBox().getSize().getValue()

    def update_label(self, obj, vobj):
        """Update the label including text size and multiplier."""
        self.font.size = vobj.TextSize.Value * vobj.ScaleMultiplier

        # Tiny additional space added to the label
        v = App.Vector(1, 0, 0)
        if obj.StraightDistance > 0:
            v = v.negative()

        v.multiply(vobj.TextSize/10)
        tsize = self.get_text_size(vobj)

        n_lines = len(obj.Text)
        total_h = tsize[1]
        height = total_h/(n_lines + 1)

        if vobj.TextAlignment == "Top":
            d = v + App.Vector(0, -height, 0)
        elif vobj.TextAlignment == "Middle":
            if n_lines == 1:
                d = v + App.Vector(0, -height/2, 0)
            else:
                d = v + App.Vector(0, -height + total_h/2, 0)
        elif vobj.TextAlignment == "Bottom":
            if n_lines == 1:
                d = v + App.Vector(0, 0, 0)
            else:
                d = v + App.Vector(0, -height + n_lines * height, 0)

        d = obj.Placement.Rotation.multVec(d)
        pos = d + obj.Placement.Base
        self.textpos.translation.setValue(pos)
        self.textpos.rotation.setValue(obj.Placement.Rotation.Q)

    def update_arrow(self, obj, vobj):
        """Update the arrow tip of the line."""
        if hasattr(self, "symbol"):
            if self.arrow.findChild(self.symbol) != -1:
                self.arrow.removeChild(self.symbol)

        s = utils.ARROW_TYPES.index(vobj.ArrowType)
        self.symbol = gui_utils.dim_symbol(s)

        if vobj.ArrowType == "Circle":
            # TODO: fix behavior of the 'Circle' marker.
            # Instead of appearing at the tip of the line
            # the 'Circle' marker appears displaced and duplicated
            # a certain distance from the tip, which is the `TargetPoint`.
            # Somehow the translation is added to the position of the tip
            # resulting in a wrong value.
            # So the arrow position is reset; nevertheless, this doesn't
            # entirely fix the issue.
            coords2 = coin.SoCoordinate3()
            coords2.point.setValues([obj.Points[-1]])
            self.arrow.addChild(coords2)
            self.arrowpos.translation.setValue((0, 0, 0))
        else:
            self.arrowpos.translation.setValue(obj.Points[-1])
        self.arrow.addChild(self.symbol)

        v1 = obj.Points[-2].sub(obj.Points[-1])

        if not DraftVecUtils.isNull(v1):
            v1.normalize()
            v2 = App.Vector(0, 0, 1)
            if round(v2.getAngle(v1), 4) in [0, round(math.pi, 4)]:
                v2 = App.Vector(0, 1, 0)

            v3 = v1.cross(v2).negative()

            _rot_mat = DraftVecUtils.getPlaneRotation(v1, v3, v2)
            q = App.Placement(_rot_mat).Rotation.Q
            self.arrowpos.rotation.setValue((q[0], q[1], q[2], q[3]))

    def draw_frame(self, obj, vobj):
        """Draw the frame around the text."""
        tsize = self.get_text_size(vobj)
        total_w = tsize[0]
        total_h = tsize[1]

        n_lines = len(obj.Text)
        height = total_h/(n_lines + 1)

        # Tiny additional space added to the label
        v = App.Vector(1, 0, 0)

        if obj.StraightDistance > 0:
            v = v.negative()
            total_w = -total_w

        v.multiply(vobj.TextSize/10)

        pts = []
        _base = obj.Placement.Base
        _pos = App.Vector(self.textpos.translation.getValue().getValue())

        # The original base position must be subtracted, otherwise the frame
        # node is displaced twice
        base = _pos - _base - v

        # Shape of the rectangle
        # (p5)p1 --------- p2
        #     |            |
        #     b            |
        #     |            |
        #     p4 --------- p3
        #
        pts.append(base + App.Vector(0, 1.07 * height, 0))
        pts.append(pts[-1] + App.Vector(total_w, 0, 0))
        pts.append(pts[-1] + App.Vector(0, -1.07 * total_h, 0))
        pts.append(pts[-1] + App.Vector(-total_w, 0, 0))
        pts.append(pts[0])

        self.fcoords.point.setValues(pts)
        self.frame.coordIndex.setValues(0,
                                        len(pts),
                                        range(len(pts)))


# Alias for compatibility with v0.18 and earlier
ViewProviderDraftLabel = ViewProviderLabel

## @}



# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/


'''
This is an example preprocessor file for the Path workbench. Its aim is to
open a gcode file, parse its contents, and create the appropriate objects
in FreeCAD.

This preprocessor will split gcode on tool changes and create one or more
PathCustom objects in the job.  Tool Change commands themselves are not
preserved. It is up to the user to create and assign appropriate tool
controllers.

Only gcodes that are supported by Path are imported. Thus things like G43
are suppressed.

Importing gcode is inherently dangerous because context cannot be safely
assumed. The user should carefully examine the resulting gcode!

Read the Path Workbench documentation to know how to create Path objects
from GCode.
'''

import os
import FreeCAD
import PathScripts.PathUtils as PathUtils
import PathScripts.PathLog as PathLog
import re
import PathScripts.PathCustom as PathCustom
import PathScripts.PathCustomGui as PathCustomGui
import PathScripts.PathOpGui as PathOpGui
from PySide import QtCore

# LEVEL = PathLog.Level.DEBUG
LEVEL = PathLog.Level.INFO
PathLog.setLevel(LEVEL, PathLog.thisModule())

if LEVEL == PathLog.Level.DEBUG:
    PathLog.trackModule(PathLog.thisModule())


# to distinguish python built-in open function from the one declared below
if open.__module__ in ['__builtin__', 'io']:
    pythonopen = open


def open(filename):
    "called when freecad opens a file."
    PathLog.track(filename)
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    insert(filename, doc.Name)


def matchToolController(op, toolnumber):
    """Try to match a tool controller in the job by number"""
    toolcontrollers = PathUtils.getToolControllers(op)
    for tc in toolcontrollers:
        if tc.ToolNumber == toolnumber:
            return tc
    return toolcontrollers[0]


def insert(filename, docname):
    "called when freecad imports a file"
    PathLog.track(filename)
    gfile = pythonopen(filename)
    gcode = gfile.read()
    gfile.close()

    # Regular expression to match tool changes in the format 'M6 Tn'
    p = re.compile('[mM]+?\s?0?6\s?T\d*\s')

    # split the gcode on tool changes
    paths = re.split('([mM]+?\s?0?6\s?T\d*\s)', gcode)

    # iterate the gcode sections and add customs for each
    toolnumber = 0

    for path in paths:

        # if the section is a tool change, extract the tool number
        m = p.match(path)
        if m:
            toolnumber = int(m.group().split('T')[-1])
            continue

        # Parse the gcode and throw away any empty lists
        gcode = parse(path)
        if len(gcode) == 0:
            continue

        # Create a custom and viewobject
        obj = PathCustom.Create("Custom")
        res = PathOpGui.CommandResources('Custom', PathCustom.Create,
                PathCustomGui.TaskPanelOpPage,
                'Path_Custom',
                QtCore.QT_TRANSLATE_NOOP('Path_Custom', 'Custom'), '', '')
        obj.ViewObject.Proxy = PathOpGui.ViewProvider(obj.ViewObject, res)
        obj.ViewObject.Proxy.setDeleteObjectsOnReject(False)

        # Set the gcode and try to match a tool controller
        obj.Gcode = gcode
        obj.ToolController = matchToolController(obj, toolnumber)

    FreeCAD.ActiveDocument.recompute()



def parse(inputstring):
    "parse(inputstring): returns a parsed output string"

    supported = ['G0', 'G00',
                 'G1', 'G01',
                 'G2', 'G02',
                 'G3', 'G03',
                 'G81', 'G82', 'G83',
                 'G90', 'G91']

    axis = ["X", "Y", "Z", "A", "B", "C", "U", "V", "W"]

    print("preprocessing...")
    PathLog.track(inputstring)
    # split the input by line
    lines = inputstring.splitlines()
    output = []
    lastcommand = None

    for lin in lines:
        # remove any leftover trailing and preceding spaces
        lin = lin.strip()

        # discard empty lines
        if not lin:
            continue

        # remove line numbers
        if lin[0].upper() in ["N"]:
            lin = lin.split(" ", 1)
            if len(lin) >= 1:
                lin = lin[1].strip()
            else:
                continue

        # Anything else not a G/M code or an axis move is ignored.
        if lin[0] not in ["G", "M", "X", "Y", "Z", "A", "B", "C", "U", "V", "W"]:
            continue

        # if the remaining line is supported, store it
        currcommand = lin.split()[0]

        if currcommand in supported:
            output.append(lin)
            lastcommand = currcommand

        # modal commands have no G or M but have axis moves. append those too.
        elif currcommand[0] in axis and lastcommand:
            output.append(lastcommand + " " + lin)

    print("done preprocessing.")
    return output


print(__name__ + " gcode preprocessor loaded.")



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import PySide
import re

__title__ = 'Generic property container to store some values.'
__author__ = 'sliptonic (Brad Collette)'
__url__ = 'https://www.freecadweb.org'
__doc__ = 'A generic container for typed properties in arbitrary categories.'

def translate(context, text, disambig=None):
    return PySide.QtCore.QCoreApplication.translate(context, text, disambig)


SupportedPropertyType = {
        'Angle'         : 'App::PropertyAngle',
        'Bool'          : 'App::PropertyBool',
        'Distance'      : 'App::PropertyDistance',
        'Enumeration'   : 'App::PropertyEnumeration',
        'File'          : 'App::PropertyFile',
        'Float'         : 'App::PropertyFloat',
        'Integer'       : 'App::PropertyInteger',
        'Length'        : 'App::PropertyLength',
        'Percent'       : 'App::PropertyPercent',
        'String'        : 'App::PropertyString',
        }

def getPropertyTypeName(o):
    for typ in SupportedPropertyType:
        if SupportedPropertyType[typ] == o:
            return typ
    raise IndexError()

class PropertyBag(object):
    '''Property container object.'''

    CustomPropertyGroups       = 'CustomPropertyGroups'
    CustomPropertyGroupDefault = 'User'

    def __init__(self, obj):
        obj.addProperty('App::PropertyStringList', self.CustomPropertyGroups, 'Base', PySide.QtCore.QT_TRANSLATE_NOOP('PathPropertyBag', 'List of custom property groups'))
        self.onDocumentRestored(obj)

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None

    def __sanitizePropertyName(self, name):
        if(len(name) == 0):
            return
        clean = name[0]
        for i in range(1, len(name)):
            if (name[i] == ' '):
                clean += name[i + 1].upper()
                i += 1
            elif(name[i - 1] != ' '):
                clean += name[i]
        return clean

    def onDocumentRestored(self, obj):
        self.obj = obj
        obj.setEditorMode(self.CustomPropertyGroups, 2)  # hide

    def getCustomProperties(self):
        '''getCustomProperties() ... Return a list of all custom properties created in this container.'''
        return [p for p in self.obj.PropertiesList if self.obj.getGroupOfProperty(p) in self.obj.CustomPropertyGroups]

    def addCustomProperty(self, propertyType, name, group=None, desc=None):
        '''addCustomProperty(propertyType, name, group=None, desc=None) ... adds a custom property and tracks its group.'''
        if desc is None:
            desc = ''
        if group is None:
            group = self.CustomPropertyGroupDefault
        groups = self.obj.CustomPropertyGroups

        name = self.__sanitizePropertyName(name)
        if not re.match("^[A-Za-z0-9_]*$", name):
            raise ValueError('Property Name can only contain letters and numbers')

        if not group in groups:
            groups.append(group)
            self.obj.CustomPropertyGroups = groups
        self.obj.addProperty(propertyType, name, group, desc)
        return name

    def refreshCustomPropertyGroups(self):
        '''refreshCustomPropertyGroups() ... removes empty property groups, should be called after deleting properties.'''
        customGroups = []
        for p in self.obj.PropertiesList:
            group = self.obj.getGroupOfProperty(p)
            if group in self.obj.CustomPropertyGroups and not group in customGroups:
                customGroups.append(group)
        self.obj.CustomPropertyGroups = customGroups


def Create(name = 'PropertyBag'):
    obj = FreeCAD.ActiveDocument.addObject('App::FeaturePython', name)
    obj.Proxy = PropertyBag(obj)
    return obj

def IsPropertyBag(obj):
    '''Returns True if the supplied object is a property container (or its Proxy).'''

    if type(obj) == PropertyBag:
        return True
    if hasattr(obj, 'Proxy'):
        return IsPropertyBag(obj.Proxy)
    return False




# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import PySide.QtCore as QtCore
import PathScripts.PathPreferences as PathPreferences


class CommandToolBitSelectorOpen:
    '''
    Command to toggle the ToolBitSelector Dock
    '''

    def __init__(self):
        pass

    def GetResources(self):
        return {'Pixmap': 'Path_ToolTable',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("PathToolBitLibrary", "ToolBit Dock"),
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("PathToolBitLibrary", "Toggle the Toolbit Dock"),
                'Accel': "P, T",
                'CmdType': "ForEdit"}

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def Activated(self):
        import PathScripts.PathToolBitLibraryGui as PathToolBitLibraryGui
        dock = PathToolBitLibraryGui.ToolBitSelector()
        dock.open()


class CommandToolBitLibraryOpen:
    '''
    Command to open ToolBitLibrary editor.
    '''

    def __init__(self):
        pass

    def GetResources(self):
        return {'Pixmap': 'Path_ToolTable',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("PathToolBitLibrary", "ToolBit Library editor"),
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("PathToolBitLibrary", "Open an editor to manage ToolBit libraries"),
                'CmdType': "ForEdit"}

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def Activated(self):
        import PathScripts.PathToolBitLibraryGui as PathToolBitLibraryGui
        library = PathToolBitLibraryGui.ToolBitLibrary()

        library.open()


if FreeCAD.GuiUp:
    FreeCADGui.addCommand('Path_ToolBitLibraryOpen', CommandToolBitLibraryOpen())
    FreeCADGui.addCommand('Path_ToolBitDock', CommandToolBitSelectorOpen())

BarList = ['Path_ToolBitDock']
MenuList = ['Path_ToolBitLibraryOpen', 'Path_ToolBitDock']

FreeCAD.Console.PrintLog("Loading PathToolBitLibraryCmd... done\n")



# ***************************************************************************
# *   Copyright (c) 2017 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver object CalculiX"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecadweb.org"

## @package SolverCalculix
#  \ingroup FEM

import glob
import os

import FreeCAD

from . import tasks
from .. import run
from .. import solverbase
from femtools import femutils

if FreeCAD.GuiUp:
    import FemGui

ANALYSIS_TYPES = ["static", "frequency", "thermomech", "check"]


def create(doc, name="SolverCalculiX"):
    return femutils.createObject(
        doc, name, Proxy, ViewProxy)


class Proxy(solverbase.Proxy):
    """The Fem::FemSolver's Proxy python type, add solver specific properties
    """

    Type = "Fem::SolverCalculix"

    def __init__(self, obj):
        super(Proxy, self).__init__(obj)
        obj.Proxy = self
        ccx_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/Ccx")
        add_attributes(obj, ccx_prefs)

    def createMachine(self, obj, directory, testmode=False):
        return run.Machine(
            solver=obj, directory=directory,
            check=tasks.Check(),
            prepare=tasks.Prepare(),
            solve=tasks.Solve(),
            results=tasks.Results(),
            testmode=testmode)

    def editSupported(self):
        return True

    def edit(self, directory):
        pattern = os.path.join(directory, "*.inp")
        FreeCAD.Console.PrintMessage("{}\n".format(pattern))
        f = glob.glob(pattern)[0]
        FemGui.open(f)

    def execute(self, obj):
        return


class ViewProxy(solverbase.ViewProxy):
    pass


# helper add properties, this is outside of the class to be able
# to use the attribute setter from framework solver and ccxtools solver
def add_attributes(obj, ccx_prefs):

    obj.addProperty(
        "App::PropertyEnumeration",
        "AnalysisType",
        "Fem",
        "Type of the analysis"
    )
    obj.AnalysisType = ANALYSIS_TYPES
    analysis_type = ccx_prefs.GetInt("AnalysisType", 0)
    obj.AnalysisType = ANALYSIS_TYPES[analysis_type]

    choices_geom_nonlinear = ["linear", "nonlinear"]
    obj.addProperty(
        "App::PropertyEnumeration",
        "GeometricalNonlinearity",
        "Fem",
        "Set geometrical nonlinearity"
    )
    obj.GeometricalNonlinearity = choices_geom_nonlinear
    nonlinear_geom = ccx_prefs.GetBool("NonlinearGeometry", False)
    if nonlinear_geom is True:
        obj.GeometricalNonlinearity = choices_geom_nonlinear[1]  # nonlinear
    else:
        obj.GeometricalNonlinearity = choices_geom_nonlinear[0]  # linear

    choices_material_nonlinear = ["linear", "nonlinear"]
    obj.addProperty(
        "App::PropertyEnumeration",
        "MaterialNonlinearity",
        "Fem",
        "Set material nonlinearity (needs geometrical nonlinearity)"
    )
    obj.MaterialNonlinearity = choices_material_nonlinear
    obj.MaterialNonlinearity = choices_material_nonlinear[0]

    obj.addProperty(
        "App::PropertyIntegerConstraint",
        "EigenmodesCount",
        "Fem",
        "Number of modes for frequency calculations"
    )
    noem = ccx_prefs.GetInt("EigenmodesCount", 10)
    obj.EigenmodesCount = (noem, 1, 100, 1)

    obj.addProperty(
        "App::PropertyFloatConstraint",
        "EigenmodeLowLimit",
        "Fem",
        "Low frequency limit for eigenmode calculations"
    )
    ell = ccx_prefs.GetFloat("EigenmodeLowLimit", 0.0)
    obj.EigenmodeLowLimit = (ell, 0.0, 1000000.0, 10000.0)

    obj.addProperty(
        "App::PropertyFloatConstraint",
        "EigenmodeHighLimit",
        "Fem",
        "High frequency limit for eigenmode calculations"
    )
    ehl = ccx_prefs.GetFloat("EigenmodeHighLimit", 1000000.0)
    obj.EigenmodeHighLimit = (ehl, 0.0, 1000000.0, 10000.0)

    help_string_IterationsThermoMechMaximum = (
        "Maximum Number of thermo mechanical iterations "
        "in each time step before stopping jobs"
    )
    obj.addProperty(
        "App::PropertyIntegerConstraint",
        "IterationsThermoMechMaximum",
        "Fem",
        help_string_IterationsThermoMechMaximum
    )
    niter = ccx_prefs.GetInt("AnalysisMaxIterations", 200)
    obj.IterationsThermoMechMaximum = niter

    obj.addProperty(
        "App::PropertyFloatConstraint",
        "TimeInitialStep",
        "Fem",
        "Initial time steps"
    )
    ini = ccx_prefs.GetFloat("AnalysisTimeInitialStep", 1.0)
    obj.TimeInitialStep = ini

    obj.addProperty(
        "App::PropertyFloatConstraint",
        "TimeEnd",
        "Fem",
        "End time analysis"
    )
    eni = ccx_prefs.GetFloat("AnalysisTime", 1.0)
    obj.TimeEnd = eni

    obj.addProperty(
        "App::PropertyBool",
        "ThermoMechSteadyState",
        "Fem",
        "Choose between steady state thermo mech or transient thermo mech analysis"
    )
    sted = ccx_prefs.GetBool("StaticAnalysis", True)
    obj.ThermoMechSteadyState = sted

    obj.addProperty(
        "App::PropertyBool",
        "IterationsControlParameterTimeUse",
        "Fem",
        "Use the user defined time incrementation control parameter"
    )
    use_non_ccx_iterations_param = ccx_prefs.GetInt("UseNonCcxIterationParam", False)
    obj.IterationsControlParameterTimeUse = use_non_ccx_iterations_param

    obj.addProperty(
        "App::PropertyBool",
        "SplitInputWriter",
        "Fem",
        "Split writing of ccx input file"
    )
    split = ccx_prefs.GetBool("SplitInputWriter", False)
    obj.SplitInputWriter = split

    ccx_default_time_incrementation_control_parameter = {
        # iteration parameter
        "I_0": 4,
        "I_R": 8,
        "I_P": 9,
        "I_C": 200,  # ccx default = 16
        "I_L": 10,
        "I_G": 400,  # ccx default = 4
        "I_S": None,
        "I_A": 200,  # ccx default = 5
        "I_J": None,
        "I_T": None,
        # cutback parameter
        "D_f": 0.25,
        "D_C": 0.5,
        "D_B": 0.75,
        "D_A": 0.85,
        "D_S": None,
        "D_H": None,
        "D_D": 1.5,
        "W_G": None}
    p = ccx_default_time_incrementation_control_parameter
    p_iter = "{0},{1},{2},{3},{4},{5},{6},{7},{8},{9}".format(
        p["I_0"],
        p["I_R"],
        p["I_P"],
        p["I_C"],
        p["I_L"],
        p["I_G"],
        "",
        p["I_A"],
        "",
        ""
    )
    p_cutb = "{0},{1},{2},{3},{4},{5},{6},{7}".format(
        p["D_f"],
        p["D_C"],
        p["D_B"],
        p["D_A"],
        "",
        "",
        p["D_D"],
        ""
    )
    obj.addProperty(
        "App::PropertyString",
        "IterationsControlParameterIter",
        "Fem",
        "User defined time incrementation iterations control parameter"
    )
    obj.IterationsControlParameterIter = p_iter
    obj.addProperty(
        "App::PropertyString",
        "IterationsControlParameterCutb",
        "Fem",
        "User defined time incrementation cutbacks control parameter"
    )
    obj.IterationsControlParameterCutb = p_cutb

    stringIterationsUserDefinedIncrementations = (
        "Set to True to switch off the ccx automatic incrementation completely "
        "(ccx parameter DIRECT). Use with care. Analysis may not converge!"
    )
    obj.addProperty(
        "App::PropertyBool",
        "IterationsUserDefinedIncrementations",
        "Fem",
        stringIterationsUserDefinedIncrementations
    )
    obj.IterationsUserDefinedIncrementations = False

    help_string_IterationsUserDefinedTimeStepLength = (
        "Set to True to use the user defined time steps. "
        "The time steps are set with TimeInitialStep and TimeEnd"
    )
    obj.addProperty(
        "App::PropertyBool",
        "IterationsUserDefinedTimeStepLength",
        "Fem",
        help_string_IterationsUserDefinedTimeStepLength
    )
    obj.IterationsUserDefinedTimeStepLength = False

    known_ccx_solver_types = [
        "default",
        "spooles",
        "iterativescaling",
        "iterativecholesky"
    ]
    obj.addProperty(
        "App::PropertyEnumeration",
        "MatrixSolverType",
        "Fem",
        "Type of solver to use"
    )
    obj.MatrixSolverType = known_ccx_solver_types
    solver_type = ccx_prefs.GetInt("Solver", 0)
    obj.MatrixSolverType = known_ccx_solver_types[solver_type]

    obj.addProperty(
        "App::PropertyBool",
        "BeamShellResultOutput3D",
        "Fem",
        "Output 3D results for 1D and 2D analysis "
    )
    dimout = ccx_prefs.GetBool("BeamShellOutput", False)
    obj.BeamShellResultOutput3D = dimout


"""
Should there be some equation object for Calculix too.

Necessarily yes! The properties GeometricalNonlinearity,
MaterialNonlinearity, ThermoMechSteadyState might be moved
to the appropriate equation.

Furthermore the material Category should not be used in writer.
See common materila object for more information. The equation
should used instead to get this information needed in writer.
"""



# ***************************************************************************
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for all annotation type objects.

This is used by many objects that show dimensions and text created on screen
through Coin (pivy).
- DimensionBase
- LinearDimension
- AngularDimension
- Label
- Text
"""
## @package draft_annotation
# \ingroup draftobjects
# \brief Provides the object code for all annotation type objects.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

from draftutils.messages import _wrn
from draftutils.translate import translate


class DraftAnnotation(object):
    """The Draft Annotation Base object.

    This class is not used directly, but inherited by all Draft annotation
    objects.

    LinearDimension through DimensionBase
    AngularDimension through DimensionBase
    Label
    Text
    """

    def __init__(self, obj, typ="Annotation"):
        self.Type = typ
        obj.Proxy = self

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored.

        Check if new properties are present after the object is restored
        in order to migrate older objects.
        """
        if hasattr(obj, "ViewObject") and obj.ViewObject:
            vobj = obj.ViewObject
            self.add_missing_properties_0v19(obj, vobj)

    def add_missing_properties_0v19(self, obj, vobj):
        """Provide missing annotation properties, if they don't exist."""
        vproperties = vobj.PropertiesList

        if 'ScaleMultiplier' not in vproperties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "General scaling factor that affects "
                                     "the annotation consistently\n"
                                     "because it scales the text, "
                                     "and the line decorations, if any,\n"
                                     "in the same proportion.")
            vobj.addProperty("App::PropertyFloat",
                             "ScaleMultiplier",
                             "Annotation",
                             _tip)
            vobj.ScaleMultiplier = 1.00

            _wrn("v0.19, " + obj.Label + ", " + translate("draft","added view property 'ScaleMultiplier'"))

        if 'AnnotationStyle' not in vproperties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Annotation style to apply to this object.\nWhen using a saved style some of the view properties will become read-only;\nthey will only be editable by changing the style through the 'Annotation style editor' tool.")
            vobj.addProperty("App::PropertyEnumeration",
                             "AnnotationStyle",
                             "Annotation",
                             _tip)
            styles = []
            for key in obj.Document.Meta.keys():
                if key.startswith("Draft_Style_"):
                    styles.append(key[12:])

            vobj.AnnotationStyle = [""] + styles

            _info = "added view property 'AnnotationStyle'"
            _wrn("v0.19, " + obj.Label + ", " + translate("draft","added view property 'ScaleMultiplier'"))

    def __getstate__(self):
        """Return a tuple of objects to save or None.

        Save the Type.
        """
        return self.Type

    def __setstate__(self, state):
        """Set the internal properties from the restored state.

        Restore the Type of the object.
        """
        if state:
            if isinstance(state, dict) and ("Type" in state):
                # During the migration of the classes
                # the 'DraftText' type was changed to 'Text' type
                if state["Type"] == "DraftText":
                    state["Type"] = "Text"
                    _wrn("v0.19, " + translate("draft","migrated 'DraftText' type to 'Text'"))
                self.Type = state["Type"]
            else:
                if state == "DraftText":
                    state = "Text"
                    _wrn("v0.19, " + translate("draft","migrated 'DraftText' type to 'Text'"))
                self.Type = state

    def execute(self, obj):
        """Execute when the object is created or recomputed.

        Does nothing.
        """
        return

    def onChanged(self, obj, prop):
        """Execute when a property is changed.

        Does nothing.
        """
        return

## @}



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to rotate shapes around a center and axis."""
## @package rotate
# \ingroup draftfuctions
# \brief Provides functions to rotate shapes around a center and axis.

## \addtogroup draftfuctions
# @{
import math

import FreeCAD as App
import DraftVecUtils
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils
import draftutils.groups as groups
import draftfunctions.join as join
import draftmake.make_line as make_line
import draftmake.make_copy as make_copy


def rotate(objectslist, angle, center=App.Vector(0,0,0),
           axis=App.Vector(0,0,1), copy=False):
    """rotate(objects,angle,[center,axis,copy])

    Rotates the objects contained in objects (that can be a list of objects
    or an object) of the given angle (in degrees) around the center, using
    axis as a rotation axis.

    Parameters
    ----------
    objectlist : list

    angle : list

    center : Base.Vector

    axis : Base.Vector
        If axis is omitted, the rotation will be around the vertical Z axis.

    copy : bool
        If copy is True, the actual objects are not moved, but copies
        are created instead.

    Return
    ----------
    The objects (or their copies) are returned.
    """
    import Part
    utils.type_check([(copy,bool)], "rotate")
    if not isinstance(objectslist,list):
        objectslist = [objectslist]

    objectslist.extend(groups.get_movable_children(objectslist))
    newobjlist = []
    newgroups = {}
    objectslist = utils.filter_objects_for_modifiers(objectslist, copy)

    for obj in objectslist:
        newobj = None
        # real_center and real_axis are introduced to take into account
        # the possibility that object is inside an App::Part
        if hasattr(obj, "getGlobalPlacement"):
            ci = obj.getGlobalPlacement().inverse().multVec(center)
            real_center = obj.Placement.multVec(ci)
            ai = obj.getGlobalPlacement().inverse().Rotation.multVec(axis)
            real_axis = obj.Placement.Rotation.multVec(ai)
        else:
            real_center = center
            real_axis = axis

        if copy:
            newobj = make_copy.make_copy(obj)
        else:
            newobj = obj
        if obj.isDerivedFrom("App::Annotation"):
            # TODO: this is very different from how move handle annotations
            # maybe we can uniform the two methods
            if axis.normalize() == App.Vector(1,0,0):
                newobj.ViewObject.RotationAxis = "X"
                newobj.ViewObject.Rotation = angle
            elif axis.normalize() == App.Vector(0,1,0):
                newobj.ViewObject.RotationAxis = "Y"
                newobj.ViewObject.Rotation = angle
            elif axis.normalize() == App.Vector(0,-1,0):
                newobj.ViewObject.RotationAxis = "Y"
                newobj.ViewObject.Rotation = -angle
            elif axis.normalize() == App.Vector(0,0,1):
                newobj.ViewObject.RotationAxis = "Z"
                newobj.ViewObject.Rotation = angle
            elif axis.normalize() == App.Vector(0,0,-1):
                newobj.ViewObject.RotationAxis = "Z"
                newobj.ViewObject.Rotation = -angle
        elif utils.get_type(obj) == "Point":
            v = App.Vector(obj.X,obj.Y,obj.Z)
            rv = v.sub(real_center)
            rv = DraftVecUtils.rotate(rv, math.radians(angle), real_axis)
            v = real_center.add(rv)
            newobj.X = v.x
            newobj.Y = v.y
            newobj.Z = v.z
        elif obj.isDerivedFrom("App::DocumentObjectGroup"):
            pass
        elif hasattr(obj,"Placement"):
            #FreeCAD.Console.PrintMessage("placement rotation\n")
            shape = Part.Shape()
            shape.Placement = obj.Placement
            shape.rotate(DraftVecUtils.tup(real_center), DraftVecUtils.tup(real_axis), angle)
            newobj.Placement = shape.Placement
        elif hasattr(obj,'Shape') and (utils.get_type(obj) not in ["WorkingPlaneProxy","BuildingPart"]):
            #think it make more sense to try first to rotate placement and later to try with shape. no?
            shape = obj.Shape.copy()
            shape.rotate(DraftVecUtils.tup(real_center), DraftVecUtils.tup(real_axis), angle)
            newobj.Shape = shape
        if copy:
            gui_utils.formatObject(newobj,obj)
        if newobj is not None:
            newobjlist.append(newobj)
        if copy:
            for p in obj.InList:
                if p.isDerivedFrom("App::DocumentObjectGroup") and (p in objectslist):
                    g = newgroups.setdefault(p.Name, App.ActiveDocument.addObject(p.TypeId, p.Name))
                    g.addObject(newobj)
                    break

    gui_utils.select(newobjlist)
    if len(newobjlist) == 1:
        return newobjlist[0]
    return newobjlist


#   Following functions are needed for SubObjects modifiers
#   implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire)


def rotate_vertex(object, vertex_index, angle, center, axis):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    points = object.Points
    points[vertex_index] = object.Placement.inverse().multVec(
        rotate_vector_from_center(
            object.Placement.multVec(points[vertex_index]),
            angle, axis, center))
    object.Points = points


rotateVertex = rotate_vertex


def rotate_vector_from_center(vector, angle, axis, center):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    rv = vector.sub(center)
    rv = DraftVecUtils.rotate(rv, math.radians(angle), axis)
    return center.add(rv)


rotateVectorFromCenter = rotate_vector_from_center


def rotate_edge(object, edge_index, angle, center, axis):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    rotateVertex(object, edge_index, angle, center, axis)
    if utils.isClosedEdge(edge_index, object):
        rotateVertex(object, 0, angle, center, axis)
    else:
        rotateVertex(object, edge_index+1, angle, center, axis)


rotateEdge = rotate_edge


def copy_rotated_edges(arguments):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    copied_edges = []
    for argument in arguments:
        copied_edges.append(copy_rotated_edge(argument[0], argument[1],
            argument[2], argument[3], argument[4]))
    join.join_wires(copied_edges)


copyRotatedEdges = copy_rotated_edges


def copy_rotated_edge(object, edge_index, angle, center, axis):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    vertex1 = rotate_vector_from_center(
        object.Placement.multVec(object.Points[edge_index]),
        angle, axis, center)
    if utils.isClosedEdge(edge_index, object):
        vertex2 = rotate_vector_from_center(
            object.Placement.multVec(object.Points[0]),
            angle, axis, center)
    else:
        vertex2 = rotate_vector_from_center(
            object.Placement.multVec(object.Points[edge_index+1]),
            angle, axis, center)
    return make_line.make_line(vertex1, vertex2)

## @}



# -*- coding: utf-8 -*
# ***************************************************************************
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# *   Bilinear interpolation code modified heavily from the interpolation   *
# *   library https://github.com/pmav99/interpolation                      *
# *   Copyright (c) 2013 by Panagiotis Mavrogiorgos                         *
# *                                                                         *
# ***************************************************************************
import FreeCAD
import FreeCADGui
import Path
import PathScripts.PathGeom as PathGeom
import PathScripts.PathLog as PathLog
import PathScripts.PathUtils as PathUtils

from PySide import QtCore, QtGui

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader
Part = LazyLoader('Part', globals(), 'Part')

"""Z Depth Correction Dressup.  This dressup takes a probe file as input and does bilinear interpolation of the Zdepths to correct for a surface which is not parallel to the milling table/bed.  The probe file should conform to the format specified by the linuxcnc G38 probe logging: 9-number coordinate consisting of XYZABCUVW http://linuxcnc.org/docs/html/gcode/g-code.html#gcode:g38
"""

LOGLEVEL = False

LOG_MODULE = PathLog.thisModule()

if LOGLEVEL:
    PathLog.setLevel(PathLog.Level.DEBUG, LOG_MODULE)
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.NOTICE, LOG_MODULE)


# Qt translation handling
def translate(context, text, disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)


movecommands = ['G1', 'G01', 'G2', 'G02', 'G3', 'G03']
rapidcommands = ['G0', 'G00']
arccommands = ['G2', 'G3', 'G02', 'G03']


class ObjectDressup:

    def __init__(self, obj):
        obj.addProperty("App::PropertyLink", "Base", "Path", QtCore.QT_TRANSLATE_NOOP("Path_DressupAxisMap", "The base path to modify"))
        obj.addProperty("App::PropertyFile", "probefile", "ProbeData", QtCore.QT_TRANSLATE_NOOP("Path_DressupZCorrect", "The point file from the surface probing."))
        obj.Proxy = self
        obj.addProperty("Part::PropertyPartShape", "interpSurface", "Path")
        obj.addProperty("App::PropertyDistance", "ArcInterpolate", "Interpolate", QtCore.QT_TRANSLATE_NOOP("Path_DressupZCorrect", "Deflection distance for arc interpolation"))
        obj.addProperty("App::PropertyDistance", "SegInterpolate", "Interpolate", QtCore.QT_TRANSLATE_NOOP("Path_DressupZCorrectp", "break segments into smaller segments of this length."))
        obj.ArcInterpolate = 0.1
        obj.SegInterpolate = 1.0

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None

    def onChanged(self, fp, prop):
        if str(prop) == "probefile":
            self._loadFile(fp, fp.probefile)

    def _bilinearInterpolate(self, surface, x, y):

        p1 = FreeCAD.Vector(x, y, 100.0)
        p2 = FreeCAD.Vector(x, y, -100.0)

        vertical_line = Part.Line(p1, p2)
        points, curves = vertical_line.intersectCS(surface)
        return points[0].Z

    def _loadFile(self, obj, filename):
        if filename == "":
            return

        f1 = open(filename, 'r')

        try:
            pointlist = []
            for line in f1.readlines():
                if line == '\n':
                    continue
                w = line.split()
                xval = round(float(w[0]), 2)
                yval = round(float(w[1]), 2)
                zval = round(float(w[2]), 2)

                pointlist.append([xval, yval, zval])
            PathLog.debug(pointlist)

            cols = list(zip(*pointlist))
            PathLog.debug("cols: {}".format(cols))
            yindex = list(sorted(set(cols[1])))
            PathLog.debug("yindex: {}".format(yindex))

            array = []
            for y in yindex:
                points = sorted([p for p in pointlist if p[1] == y])
                inner = []
                for p in points:
                    inner.append(FreeCAD.Vector(p[0], p[1], p[2]))
                array.append(inner)

            intSurf = Part.BSplineSurface()
            intSurf.interpolate(array)

            obj.interpSurface = intSurf.toShape()
        except Exception:
            raise ValueError("File does not contain appropriate point data")

    def execute(self, obj):

        sampleD = obj.SegInterpolate.Value
        curveD = obj.ArcInterpolate.Value

        if obj.interpSurface.isNull():  # No valid probe data.  return unchanged path
            obj.Path = obj.Base.Path
            return

        surface = obj.interpSurface.toNurbs().Faces[0].Surface

        if obj.Base:
            if obj.Base.isDerivedFrom("Path::Feature"):
                if obj.Base.Path:
                    if obj.Base.Path.Commands:
                        pathlist = obj.Base.Path.Commands

                        newcommandlist = []
                        currLocation = {'X': 0, 'Y': 0, 'Z': 0, 'F': 0}

                        for c in pathlist:
                            PathLog.debug(c)
                            PathLog.debug("     curLoc:{}".format(currLocation))
                            newparams = dict(c.Parameters)
                            zval = newparams.get("Z", currLocation['Z'])
                            if c.Name in movecommands:
                                curVec = FreeCAD.Vector(currLocation['X'], currLocation['Y'], currLocation['Z'])
                                arcwire = PathGeom.edgeForCmd(c, curVec)
                                if arcwire is None:
                                    continue
                                if c.Name in arccommands:
                                    pointlist = arcwire.discretize(Deflection=curveD)
                                else:
                                    disc_number = int(arcwire.Length / sampleD)
                                    if disc_number > 1:
                                        pointlist = arcwire.discretize(Number=int(arcwire.Length / sampleD))
                                    else:
                                        pointlist = [v.Point for v in arcwire.Vertexes]
                                for point in pointlist:
                                    offset = self._bilinearInterpolate(surface, point.x, point.y)
                                    newcommand = Path.Command("G1", {'X': point.x, 'Y': point.y, 'Z': point.z + offset})
                                    newcommandlist.append(newcommand)
                                    currLocation.update(newcommand.Parameters)
                                    currLocation['Z'] = zval

                            else:
                                # Non Feed Command
                                newcommandlist.append(c)
                                currLocation.update(c.Parameters)
                        path = Path.Path(newcommandlist)
                        obj.Path = path


class TaskPanel:

    def __init__(self, obj):
        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/ZCorrectEdit.ui")
        FreeCAD.ActiveDocument.openTransaction(translate("Path_DressupZCorrect", "Edit Z Correction Dress-up"))
        self.interpshape = FreeCAD.ActiveDocument.addObject("Part::Feature", "InterpolationSurface")
        self.interpshape.Shape = obj.interpSurface
        self.interpshape.ViewObject.Transparency = 60
        self.interpshape.ViewObject.ShapeColor = (1.00000, 1.00000, 0.01961)
        self.interpshape.ViewObject.Selectable = False
        stock = PathUtils.findParentJob(obj).Stock
        self.interpshape.Placement.Base.z = stock.Shape.BoundBox.ZMax

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def accept(self):
        self.getFields()
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.removeObject(self.interpshape.Name)
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()
        FreeCAD.ActiveDocument.recompute()

    def getFields(self):
        self.obj.Proxy.execute(self.obj)

    def updateUI(self):

        if PathLog.getLevel(LOG_MODULE) == PathLog.Level.DEBUG:
            for obj in FreeCAD.ActiveDocument.Objects:
                if obj.Name.startswith('Shape'):
                    FreeCAD.ActiveDocument.removeObject(obj.Name)
            print('object name %s' % self.obj.Name)
            if hasattr(self.obj.Proxy, "shapes"):
                PathLog.info("showing shapes attribute")
                for shapes in self.obj.Proxy.shapes.itervalues():
                    for shape in shapes:
                        Part.show(shape)
            else:
                PathLog.info("no shapes attribute found")

    def updateModel(self):
        self.getFields()
        self.updateUI()
        FreeCAD.ActiveDocument.recompute()

    def setFields(self):
        self.form.ProbePointFileName.setText(self.obj.probefile)

        self.updateUI()

    def open(self):
        pass

    def setupUi(self):
        self.setFields()
        # now that the form is filled, setup the signal handlers
        self.form.ProbePointFileName.editingFinished.connect(self.updateModel)
        self.form.SetProbePointFileName.clicked.connect(self.SetProbePointFileName)

    def SetProbePointFileName(self):
        filename = QtGui.QFileDialog.getOpenFileName(self.form, translate("Path_Probe", "Select Probe Point File"), None, translate("Path_Probe", "All Files (*.*)"))
        if filename and filename[0]:
            self.obj.probefile = str(filename[0])
            self.setFields()


class ViewProviderDressup:

    def __init__(self, vobj):
        vobj.Proxy = self

    def attach(self, vobj):
        self.obj = vobj.Object
        if self.obj and self.obj.Base:
            for i in self.obj.Base.InList:
                if hasattr(i, "Group"):
                    group = i.Group
                    for g in group:
                        if g.Name == self.obj.Base.Name:
                            group.remove(g)
                    i.Group = group
        return

    def claimChildren(self):
        return [self.obj.Base]

    def setEdit(self, vobj, mode=0):
        FreeCADGui.Control.closeDialog()
        panel = TaskPanel(vobj.Object)
        FreeCADGui.Control.showDialog(panel)
        panel.setupUi()
        return True

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None

    def onDelete(self, arg1=None, arg2=None):
        '''this makes sure that the base operation is added back to the project and visible'''
        FreeCADGui.ActiveDocument.getObject(arg1.Object.Base.Name).Visibility = True
        job = PathUtils.findParentJob(arg1.Object)
        job.Proxy.addOperation(arg1.Object.Base)
        arg1.Object.Base = None
        return True


class CommandPathDressup:

    def GetResources(self):
        return {'Pixmap': 'Path_Dressup',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("Path_DressupZCorrect", "Z Depth Correction Dress-up"),
                'Accel': "",
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("Path_DressupZCorrect", "Use Probe Map to correct Z depth")}

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                        return True
        return False

    def Activated(self):

        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            FreeCAD.Console.PrintError(translate("Path_Dressup", "Please select one path object\n"))
            return
        if not selection[0].isDerivedFrom("Path::Feature"):
            FreeCAD.Console.PrintError(translate("Path_Dressup", "The selected object is not a path\n"))
            return
        if selection[0].isDerivedFrom("Path::FeatureCompoundPython"):
            FreeCAD.Console.PrintError(translate("Path_Dressup", "Please select a Path object"))
            return

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction(translate("Path_DressupZCorrect", "Create Dress-up"))
        FreeCADGui.addModule("PathScripts.PathDressupZCorrect")
        FreeCADGui.addModule("PathScripts.PathUtils")
        FreeCADGui.doCommand('obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", "ZCorrectDressup")')
        FreeCADGui.doCommand('PathScripts.PathDressupZCorrect.ObjectDressup(obj)')
        FreeCADGui.doCommand('obj.Base = FreeCAD.ActiveDocument.' + selection[0].Name)
        FreeCADGui.doCommand('PathScripts.PathDressupZCorrect.ViewProviderDressup(obj.ViewObject)')
        FreeCADGui.doCommand('PathScripts.PathUtils.addToJob(obj)')
        FreeCADGui.doCommand('Gui.ActiveDocument.getObject(obj.Base.Name).Visibility = False')
        FreeCADGui.doCommand('obj.ViewObject.Document.setEdit(obj.ViewObject, 0)')
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand('Path_DressupZCorrect', CommandPathDressup())

FreeCAD.Console.PrintLog("Loading PathDressup... done\n")



# -*- coding: utf-8 -*-
#***************************************************************************
#*                                                                         *
#*   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

import os
import re
import shutil
import sys

from PySide import QtCore, QtGui

import FreeCAD
if FreeCAD.GuiUp:
    import FreeCADGui

import addonmanager_utilities as utils
from addonmanager_utilities import translate  # this needs to be as is for pylupdate
from addonmanager_macro import Macro

have_git = False
try:
    import git
    have_git = True
except ImportError:
    pass

have_zip = False
try:
    import zipfile
    have_zip = True
except ImportError:
    pass

have_markdown = False
try:
    import markdown
    have_markdown = True
except ImportError:
    pass

try:
    import StringIO as io
    _stringio = io.StringIO
except ImportError:  # StringIO is not available with python3
    import io
    _stringio = io.BytesIO

#  @package AddonManager_workers
#  \ingroup ADDONMANAGER
#  \brief Multithread workers for the addon manager
#  @{

# Blacklisted addons
macros_blacklist = []

# These addons will print an additional message informing the user
obsolete = []

# These addons will print an additional message informing the user Python2 only
py2only = []

NOGIT = False  # for debugging purposes, set this to True to always use http downloads

NOMARKDOWN = False  # for debugging purposes, set this to True to disable Markdown lib

url_to_branch_map = {} # Added to 0.19.3 to support branches with names other than "master"

"""Multithread workers for the Addon Manager"""


class UpdateWorker(QtCore.QThread):
    """This worker updates the list of available workbenches"""

    info_label = QtCore.Signal(str)
    addon_repo = QtCore.Signal(object)
    progressbar_show = QtCore.Signal(bool)
    done = QtCore.Signal()

    def __init__(self):

        QtCore.QThread.__init__(self)

    def run(self):
        "populates the list of addons"

        self.progressbar_show.emit(True)

        # update info lists
        global obsolete, macros_blacklist, py2only
        u = utils.urlopen("https://raw.githubusercontent.com/FreeCAD/FreeCAD-addons/master/addonflags.json")
        if u:
            p = u.read()
            if sys.version_info.major >= 3 and isinstance(p, bytes):
                p = p.decode("utf-8")
            u.close()
            hit = re.findall(r'"obsolete"[^\{]*?{[^\{]*?"Mod":\[(?P<obsolete>[^\[\]]+?)\]}',
                             p.replace("\n", "").replace(" ", ""))
            if hit:
                obsolete = hit[0].replace('"', "").split(",")
            hit = re.findall(r'"blacklisted"[^\{]*?{[^\{]*?"Macro":\[(?P<blacklisted>[^\[\]]+?)\]}',
                             p.replace("\n", "").replace(" ", ""))
            if hit:
                macros_blacklist = hit[0].replace('"', "").split(",")
            hit = re.findall(r'"py2only"[^\{]*?{[^\{]*?"Mod":\[(?P<py2only>[^\[\]]+?)\]}',
                             p.replace("\n", "").replace(" ", ""))
            if hit:
                py2only = hit[0].replace('"', "").split(",")
        else:
            print("Debug: addon_flags.json not found")

        u = utils.urlopen("https://raw.githubusercontent.com/FreeCAD/FreeCAD-addons/master/.gitmodules")
        if not u:
            self.progressbar_show.emit(False)
            self.done.emit()
            self.stop = True
            return
        p = u.read()
        if sys.version_info.major >= 3 and isinstance(p, bytes):
            p = p.decode("utf-8")
        u.close()
        p = re.findall((r'(?m)\[submodule\s*"(?P<name>.*)"\]\s*'
                        r"path\s*=\s*(?P<path>.+)\s*"
                        r"url\s*=\s*(?P<url>https?://.*)\s*"
                        r"(branch\s*=\s*(?P<branch>.*)\s*)?"), p)
        basedir = FreeCAD.getUserAppDataDir()
        moddir = basedir + os.sep + "Mod"
        repos = []
        # querying official addons
        for name, path, url, _, branch in p:
            self.info_label.emit(name)
            url = url.split(".git")[0]
            addondir = moddir + os.sep + name
            if os.path.exists(addondir) and os.listdir(addondir):
                # make sure the folder exists and it contains files!
                state = 1
            else:
                state = 0
            repos.append([name, url, state])
            if branch is None or len(branch) == 0:
                 branch = "master"
            url_to_branch_map[url] = branch
        # querying custom addons
        customaddons = (FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
                        .GetString("CustomRepositories", "").split("\n"))
        for url in customaddons:
            if url:
                if " " in url:
                    addon_and_branch = addon.split(" ")
                    url = addon_and_branch[0]
                    branch = addon_and_branch[1]
                else:
                    branch = "master"
                name = url.split("/")[-1]
                if name.lower().endswith(".git"):
                    name = name[:-4]
                addondir = moddir + os.sep + name
                if not os.path.exists(addondir):
                    state = 0
                else:
                    state = 1
                repos.append([name, url, state])
                url_to_branch_map[url] = branch
        if not repos:
            self.info_label.emit(translate("AddonsInstaller", "Unable to download addon list."))
        else:
            repos = sorted(repos, key=lambda s: s[0].lower())
            for repo in repos:
                self.addon_repo.emit(repo)
            self.info_label.emit(translate("AddonsInstaller", "Workbenches list was updated."))
        self.progressbar_show.emit(False)
        self.done.emit()
        self.stop = True


class InfoWorker(QtCore.QThread):
    """This worker retrieves the description text of a workbench"""

    addon_repos = QtCore.Signal(object)

    def __init__(self):

        QtCore.QThread.__init__(self)

    def run(self):

        i = 0
        for repo in self.repos:
            url = repo[1]
            u = utils.urlopen(url)
            if not u:
                self.stop = True
                return
            p = u.read()
            if sys.version_info.major >= 3 and isinstance(p, bytes):
                p = p.decode("utf-8")
            u.close()
            desc = re.findall('<meta property="og:description" content="(.*?)"', p)
            if desc:
                desc = desc[0]
            else:
                desc = "Unable to retrieve addon description"
            self.repos[i].append(desc)
            i += 1
            self.addon_repos.emit(self.repos)
        self.stop = True


class CheckWBWorker(QtCore.QThread):
    """This worker checks for available updates for all workbenches"""

    enable = QtCore.Signal(int)
    mark = QtCore.Signal(str)
    addon_repos = QtCore.Signal(object)

    def __init__(self, repos):

        QtCore.QThread.__init__(self)
        self.repos = repos

    def run(self):

        if NOGIT or not have_git:
            self.stop = True
            return
        basedir = FreeCAD.getUserAppDataDir()
        moddir = basedir + os.sep + "Mod"
        upds = []
        gitpython_warning = False
        for repo in self.repos:
            if repo[2] == 1:  # installed
                # print("Checking for updates for", repo[0])
                clonedir = moddir + os.sep + repo[0]
                if os.path.exists(clonedir):
                    # mark as already installed AND already checked for updates
                    self.repos[self.repos.index(repo)][2] = 2
                    if not os.path.exists(clonedir + os.sep + ".git"):
                        # Repair addon installed with raw download
                        bare_repo = git.Repo.clone_from(repo[1], clonedir + os.sep + ".git", bare=True)
                        try:
                            with bare_repo.config_writer() as cw:
                                cw.set("core", "bare", False)
                        except AttributeError:
                            if not gitpython_warning:
                                FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                                       "Outdated GitPython detected, "
                                                                       "consider upgrading with pip.") + "\n")
                                gitpython_warning = True
                            cw = bare_repo.config_writer()
                            cw.set("core", "bare", False)
                            del cw
                        repo = git.Repo(clonedir)
                        repo.head.reset("--hard")
                    gitrepo = git.Git(clonedir)
                    try:
                        gitrepo.fetch()
                    except Exception:
                        print("AddonManager: Unable to fetch git updates for repo", repo[0])
                    else:
                        try:
                            if "git pull" in gitrepo.status():
                                self.mark.emit(repo[0])
                                upds.append(repo[0])
                                # mark as already installed AND already checked for updates AND update available
                                self.repos[self.repos.index(repo)][2] = 3
                        except stderr:
                            FreeCAD.Console.PrintWarning("AddonManager - " + repo[0] + " git status"
                                                         " fatal: this operation must be run in a work tree \n")
        self.addon_repos.emit(self.repos)
        self.enable.emit(len(upds))
        self.stop = True


class FillMacroListWorker(QtCore.QThread):
    """This worker populates the list of macros"""

    add_macro_signal = QtCore.Signal(Macro)
    info_label_signal = QtCore.Signal(str)
    progressbar_show = QtCore.Signal(bool)

    def __init__(self, repo_dir):

        QtCore.QThread.__init__(self)
        self.repo_dir = repo_dir
        self.macros = []

    def run(self):
        """Populates the list of macros"""

        self.retrieve_macros_from_git()
        self.retrieve_macros_from_wiki()
        [self.add_macro_signal.emit(m) for m in sorted(self.macros, key=lambda m: m.name.lower())]
        if self.macros:
            self.info_label_signal.emit(translate("AddonsInstaller", "List of macros successfully retrieved."))
        self.progressbar_show.emit(False)
        self.stop = True

    def retrieve_macros_from_git(self):
        """Retrieve macros from FreeCAD-macros.git

        Emits a signal for each macro in
        https://github.com/FreeCAD/FreeCAD-macros.git
        """

        if not have_git:
            self.info_label_signal.emit("GitPython not installed! Cannot retrieve macros from Git")
            FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                   "GitPython not installed! Cannot retrieve macros from git")+"\n")
            return

        self.info_label_signal.emit("Downloading list of macros from git...")
        try:
            git.Repo.clone_from("https://github.com/FreeCAD/FreeCAD-macros.git", self.repo_dir)
        except Exception:
            FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                   "Something went wrong with the Git Macro Retrieval, "
                                                   "possibly the Git executable is not in the path") + "\n")
        for dirpath, _, filenames in os.walk(self.repo_dir):
            if ".git" in dirpath:
                continue
            for filename in filenames:
                if filename.lower().endswith(".fcmacro"):
                    macro = Macro(filename[:-8])  # Remove ".FCMacro".
                    macro.on_git = True
                    macro.src_filename = os.path.join(dirpath, filename)
                    self.macros.append(macro)

    def retrieve_macros_from_wiki(self):
        """Retrieve macros from the wiki

        Read the wiki and emit a signal for each found macro.
        Reads only the page https://www.freecadweb.org/wiki/Macros_recipes
        """

        self.info_label_signal.emit("Downloading list of macros from the FreeCAD wiki...")
        self.progressbar_show.emit(True)
        u = utils.urlopen("https://www.freecadweb.org/wiki/Macros_recipes")
        if not u:
            FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                   "Appears to be an issue connecting to the Wiki, "
                                                   "therefore cannot retrieve Wiki macro list at this time") + "\n")
            return
        p = u.read()
        u.close()
        if sys.version_info.major >= 3 and isinstance(p, bytes):
            p = p.decode("utf-8")
        macros = re.findall('title="(Macro.*?)"', p)
        macros = [mac for mac in macros if ("translated" not in mac)]
        for mac in macros:
            macname = mac[6:]  # Remove "Macro ".
            macname = macname.replace("&amp;", "&")
            if (macname not in macros_blacklist) and ("recipes" not in macname.lower()):
                macro = Macro(macname)
                macro.on_wiki = True
                self.macros.append(macro)


class ShowWorker(QtCore.QThread):
    """This worker retrieves info of a given workbench"""

    info_label = QtCore.Signal(str)
    addon_repos = QtCore.Signal(object)
    progressbar_show = QtCore.Signal(bool)

    def __init__(self, repos, idx):

        # repos is a list of [name, url, installbit, descr]
        #   name      : Addon name
        #   url       : Addon repository location
        #   installbit: 0 = Addon is not installed
        #               1 = Addon is installed
        #               2 = Addon is installed and checked for available updates (none pending)
        #               3 = Addon is installed and has a pending update
        #   descr     : Addon description

        QtCore.QThread.__init__(self)
        self.repos = repos
        self.idx = idx

    def run(self):

        self.progressbar_show.emit(True)
        self.info_label.emit(translate("AddonsInstaller", "Retrieving description..."))

        if len(self.repos[self.idx]) == 4:
            desc = self.repos[self.idx][3]
        else:
            u = None
            url = self.repos[self.idx][1]
            if url in url_to_branch_map:
                branch = url_to_branch_map[url]
            else:
                branch = "master"
            self.info_label.emit(translate("AddonsInstaller", "Retrieving info from") + " " + str(url))
            desc = ""
            regex = utils.get_readme_regex(url)
            if regex:
                # extract readme from html via regex
                readmeurl = utils.get_readme_html_url(url,branch)
                if not readmeurl:
                    print("Debug: README not found for", url)
                u = utils.urlopen(readmeurl)
                if not u:
                    print("Debug: README not found at", readmeurl)
                u = utils.urlopen(readmeurl)
                if u:
                    p = u.read()
                    if sys.version_info.major >= 3 and isinstance(p, bytes):
                        p = p.decode("utf-8")
                    u.close()
                    readme = re.findall(regex, p, flags=re.MULTILINE | re.DOTALL)
                    if readme:
                        desc = readme[0]
                else:
                    print("Debug: README not found at", readmeurl)
            else:
                # convert raw markdown using lib
                readmeurl = utils.get_readme_url(url,branch)
                if not readmeurl:
                    print("Debug: README not found for", url)
                u = utils.urlopen(readmeurl)
                if u:
                    p = u.read()
                    if sys.version_info.major >= 3 and isinstance(p, bytes):
                        p = p.decode("utf-8")
                    u.close()
                    desc = utils.fix_relative_links(p, readmeurl.rsplit("/README.md")[0])
                    if have_markdown and not NOMARKDOWN:
                        desc = markdown.markdown(desc, extensions=["md_in_html"])
                    else:
                        message = """
<div style="width: 100%; text-align:center;background: #91bbe0;">
    <strong style="color: #FFFFFF;">
"""
                        message += translate("AddonsInstaller", "Raw markdown displayed")
                        message += "</strong><br/><br/>"
                        message += translate("AddonsInstaller", "Python Markdown library is missing.")
                        message += "<br/></div><hr/><pre>" + desc + "</pre>"
                        desc = message
                else:
                    print("Debug: README not found at", readmeurl)
            if desc == "":
                # fall back to the description text
                u = utils.urlopen(url)
                if not u:
                    self.progressbar_show.emit(False)
                    self.stop = True
                    return
                p = u.read()
                if sys.version_info.major >= 3 and isinstance(p, bytes):
                    p = p.decode("utf-8")
                u.close()
                descregex = utils.get_desc_regex(url)
                if descregex:
                    desc = re.findall(descregex, p)
                    if desc:
                        desc = desc[0]
            if not desc:
                desc = "Unable to retrieve addon description"
            self.repos[self.idx].append(desc)
            self.addon_repos.emit(self.repos)
        # Addon is installed so lets check if it has an update
        if self.repos[self.idx][2] == 1:
            upd = False
            # checking for updates
            if not NOGIT and have_git:
                repo = self.repos[self.idx]
                clonedir = FreeCAD.getUserAppDataDir() + os.sep + "Mod" + os.sep + repo[0]
                if os.path.exists(clonedir):
                    if not os.path.exists(clonedir + os.sep + ".git"):
                        # Repair addon installed with raw download
                        bare_repo = git.Repo.clone_from(repo[1], clonedir + os.sep + ".git", bare=True)
                        try:
                            with bare_repo.config_writer() as cw:
                                cw.set("core", "bare", False)
                        except AttributeError:
                            FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                                   "Outdated GitPython detected, "
                                                                   "consider upgrading with pip.") + "\n")
                            cw = bare_repo.config_writer()
                            cw.set("core", "bare", False)
                            del cw
                        repo = git.Repo(clonedir)
                        repo.head.reset("--hard")
                    gitrepo = git.Git(clonedir)
                    gitrepo.fetch()
                    if "git pull" in gitrepo.status():
                        upd = True
            # If there is an update pending, lets user know via the UI
            if upd:
                message = """
<div style="width: 100%;text-align: center;background: #75AFFD;">
    <br/>
    <strong style="background: #397FF7;color: #FFFFFF;">
"""
                message += translate("AddonsInstaller", "An update is available for this addon.")
                message += "</strong><br/></div><hr/>" + desc + '<br/><br/>Addon repository: <a href="'
                message += self.repos[self.idx][1] + '">' + self.repos[self.idx][1] + "</a>"
                # mark as already installed AND already checked for updates AND update is available
                self.repos[self.idx][2] = 3
            # If there isn't, indicate that this addon is already installed
            else:
                message = """
<div style="width: 100%;text-align: center;background: #C1FEB2;">
    <br/>
    <strong style="background: #00B629;color: #FFFFFF;">
"""
                message += translate("AddonsInstaller", "This addon is already installed.")
                message += "</strong><br/></div><hr/>" + desc
                message += '<br/><br/>Addon repository: <a href="'
                message += self.repos[self.idx][1] + '">' + self.repos[self.idx][1] + "</a>"
                self.repos[self.idx][2] = 2  # mark as already installed AND already checked for updates
            # Let the user know the install path for this addon
            message += "<br/>" + translate("AddonInstaller", "Installed location") + ": "
            message += FreeCAD.getUserAppDataDir() + os.sep + "Mod" + os.sep + self.repos[self.idx][0]
            self.addon_repos.emit(self.repos)
        elif self.repos[self.idx][2] == 2:
            message = """
<div style="width: 100%;text-align: center;background: #C1FEB2;">
    <br/>
    <strong style="background: #00B629;color: #FFFFFF;">
"""
            message += translate("AddonsInstaller", "This addon is already installed.")
            message += "</strong><br></div><hr/>" + desc
            message += '<br/><br/>Addon repository: <a href="'
            message += self.repos[self.idx][1] + '">' + self.repos[self.idx][1] + "</a>"
            message += "<br/>" + translate("AddonInstaller", "Installed location") + ": "
            message += FreeCAD.getUserAppDataDir() + os.sep + "Mod" + os.sep + self.repos[self.idx][0]
        elif self.repos[self.idx][2] == 3:
            message = """
<div style="width: 100%;text-align: center;background: #75AFFD;">
    <br/>
    <strong style="background: #397FF7;color: #FFFFFF;">
"""
            message += translate("AddonsInstaller", "An update is available for this addon.")
            message += "</strong><br/></div><hr/>" + desc + '<br/><br/>Addon repository: <a href="'
            message += self.repos[self.idx][1] + '">' + self.repos[self.idx][1] + "</a>"
            message += "<br/>" + translate("AddonInstaller", "Installed location") + ": "
            message += FreeCAD.getUserAppDataDir() + os.sep + "Mod" + os.sep + self.repos[self.idx][0]
        else:
            message = desc + '<br/><br/>Addon repository: <a href="'
            message += self.repos[self.idx][1] + '">' + self.repos[self.idx][1] + '</a>'

        # If the Addon is obsolete, let the user know through the Addon UI
        if self.repos[self.idx][0] in obsolete:
            message = """
<div style="width: 100%; text-align:center; background: #FFB3B3;">
    <strong style="color: #FFFFFF; background: #FF0000;">
"""
            message += translate("AddonsInstaller", "This addon is marked as obsolete") + "</strong><br/><br/>"
            message += translate("AddonsInstaller",
                                 "This usually means it is no longer maintained, "
                                 "and some more advanced addon in this list "
                                 "provides the same functionality.") + "<br/></div><hr/>" + desc

        # If the Addon is Python 2 only, let the user know through the Addon UI
        if self.repos[self.idx][0] in py2only:
            message = """
<div style="width: 100%; text-align:center; background: #ffe9b3;">
    <strong style="color: #FFFFFF; background: #ff8000;">
"""
            message += translate("AddonsInstaller", "This addon is marked as Python 2 Only") + "</strong><br/><br/>"
            message += translate("AddonsInstaller",
                                 "This workbench may no longer be maintained and "
                                 "installing it on a Python 3 system will more than "
                                 "likely result in errors at startup or while in use.")
            message += "<br/></div><hr/>" + desc

        self.info_label.emit(message)
        self.progressbar_show.emit(False)
        self.mustLoadImages = True
        label = self.loadImages(message, self.repos[self.idx][1], self.repos[self.idx][0])
        if label:
            self.info_label.emit(label)
        self.stop = True

    def stopImageLoading(self):
        "this stops the image loading process and allow the thread to terminate earlier"

        self.mustLoadImages = False

    def loadImages(self, message, url, wbName):
        "checks if the given page contains images and downloads them"

        # QTextBrowser cannot display online images. So we download them
        # here, and replace the image link in the html code with the
        # downloaded version

        imagepaths = re.findall("<img.*?src=\"(.*?)\"", message)
        if imagepaths:
            storedimages = []
            store = os.path.join(FreeCAD.getUserAppDataDir(), "AddonManager", "Images")
            if not os.path.exists(store):
                os.makedirs(store)
            for path in imagepaths:
                if not self.mustLoadImages:
                    return None
                origpath = path
                if "?" in path:
                    # remove everything after the ?
                    path = path.split("?")[0]
                if not path.startswith("http"):
                    path = utils.getserver(url) + path
                name = path.split("/")[-1]
                if name and path.startswith("http"):
                    storename = os.path.join(store, name)
                    if len(storename) >= 260:
                        remainChars = 259 - (len(store) + len(wbName) + 1)
                        storename = os.path.join(store, wbName+name[-remainChars:])
                    if not os.path.exists(storename):
                        try:
                            u = utils.urlopen(path)
                            imagedata = u.read()
                            u.close()
                        except Exception:
                            print("AddonManager: Debug: Error retrieving image from", path)
                        else:
                            f = open(storename, "wb")
                            f.write(imagedata)
                            f.close()

                            # resize the image to 300x300px if needed
                            img = QtGui.QImage(storename)
                            if (img.width() > 300) or (img.height() > 300):
                                pix = QtGui.QPixmap()
                                pix = pix.fromImage(img.scaled(300, 300,
                                                               QtCore.Qt.KeepAspectRatio,
                                                               QtCore.Qt.FastTransformation))
                                pix.save(storename, "jpeg", 100)
                    message = message.replace("src=\""+origpath, "src=\"file:///"+storename.replace("\\", "/"))
            # print(message)
            return message
        return None


class GetMacroDetailsWorker(QtCore.QThread):
    """Retrieve the macro details for a macro"""

    info_label = QtCore.Signal(str)
    progressbar_show = QtCore.Signal(bool)

    def __init__(self, macro):

        QtCore.QThread.__init__(self)
        self.macro = macro

    def run(self):

        self.progressbar_show.emit(True)
        self.info_label.emit(translate("AddonsInstaller", "Retrieving description..."))
        if not self.macro.parsed and self.macro.on_git:
            self.info_label.emit(translate("AddonsInstaller", "Retrieving info from git"))
            self.macro.fill_details_from_file(self.macro.src_filename)
        if not self.macro.parsed and self.macro.on_wiki:
            self.info_label.emit(translate("AddonsInstaller", "Retrieving info from wiki"))
            mac = self.macro.name.replace(" ", "_")
            mac = mac.replace("&", "%26")
            mac = mac.replace("+", "%2B")
            url = "https://www.freecadweb.org/wiki/Macro_" + mac
            self.macro.fill_details_from_wiki(url)
        if self.macro.is_installed():
            already_installed_msg = ('<strong style=\"background: #00B629;\">'
                                     + translate("AddonsInstaller", "This macro is already installed.")
                                     + '</strong><br>')
        else:
            already_installed_msg = ""
        message = (already_installed_msg
                   + "<h1>"+self.macro.name+"</h1>"
                   + self.macro.desc
                   + "<br/><br/>Macro location: <a href=\""
                   + self.macro.url
                   + "\">"
                   + self.macro.url
                   + "</a>")
        self.info_label.emit(message)
        self.progressbar_show.emit(False)
        self.stop = True


class InstallWorker(QtCore.QThread):
    "This worker installs a workbench"

    info_label = QtCore.Signal(str)
    progressbar_show = QtCore.Signal(bool)
    mark_recompute = QtCore.Signal(str)

    def __init__(self, repos, idx):

        QtCore.QThread.__init__(self)
        self.idx = idx
        self.repos = repos

    def run(self):
        "installs or updates the selected addon"

        if not have_git:
            self.info_label.emit("GitPython not found.")
            FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                   "GitPython not found. Using standard download instead.") + "\n")
            if not have_zip:
                self.info_label.emit("no zip support.")
                FreeCAD.Console.PrintError(translate("AddonsInstaller",
                                                     "Your version of python doesn't appear to support ZIP "
                                                     "files. Unable to proceed.") + "\n")
                return

        if not isinstance(self.idx, list):
            self.idx = [self.idx]
        for idx in self.idx:
            if idx < 0:
                return
            if not self.repos:
                return
            basedir = FreeCAD.getUserAppDataDir()
            moddir = basedir + os.sep + "Mod"
            if not os.path.exists(moddir):
                os.makedirs(moddir)
            clonedir = moddir + os.sep + self.repos[idx][0]
            self.progressbar_show.emit(True)
            if os.path.exists(clonedir):
                self.info_label.emit("Updating module...")
                if sys.version_info.major > 2 and str(self.repos[idx][0]) in py2only:
                    FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                           "User requested updating a Python 2 workbench on "
                                                           "a system running Python 3 - ") +
                                                 str(self.repos[idx][0])+"\n")
                if have_git:
                    if not os.path.exists(clonedir + os.sep + ".git"):
                        # Repair addon installed with raw download
                        bare_repo = git.Repo.clone_from(self.repos[idx][1], clonedir + os.sep + ".git", bare=True)
                        try:
                            with bare_repo.config_writer() as cw:
                                cw.set("core", "bare", False)
                        except AttributeError:
                            FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                                   "Outdated GitPython detected, consider "
                                                                   "upgrading with pip.") + "\n")
                            cw = bare_repo.config_writer()
                            cw.set("core", "bare", False)
                            del cw
                        repo = git.Repo(clonedir)
                        repo.head.reset("--hard")
                    repo = git.Git(clonedir)
                    try:
                        answer = repo.pull() + "\n\n" + translate("AddonsInstaller",
                                                                  "Workbench successfully updated. "
                                                                  "Please restart FreeCAD to apply the changes.")
                    except Exception:
                        print("Error updating module", self.repos[idx][1], " - Please fix manually")
                        answer = repo.status()
                        print(answer)
                    else:
                        # Update the submodules for this repository
                        repo_sms = git.Repo(clonedir)
                        for submodule in repo_sms.submodules:
                            submodule.update(init=True, recursive=True)
                else:
                    answer = self.download(self.repos[idx][1], clonedir) + "\n\n"
                    answer += translate("AddonsInstaller",
                                        "Workbench successfully updated. Please "
                                        "restart FreeCAD to apply the changes.")
            else:
                self.info_label.emit("Checking module dependencies...")
                depsok, answer = self.check_dependencies(self.repos[idx][1])
                if depsok:
                    if sys.version_info.major > 2 and str(self.repos[idx][0]) in py2only:
                        FreeCAD.Console.PrintWarning(translate("AddonsInstaller",
                                                               "User requested installing a Python 2 "
                                                               "workbench on a system running Python 3 - ") +
                                                     str(self.repos[idx][0]) + "\n")
                    if have_git:
                        self.info_label.emit("Cloning module...")
                        repo = git.Repo.clone_from(self.repos[idx][1], clonedir)

                        # Make sure to clone all the submodules as well
                        if repo.submodules:
                            repo.submodule_update(recursive=True)
                    else:
                        self.info_label.emit("Downloading module...")
                        self.download(self.repos[idx][1], clonedir)
                    answer = translate("AddonsInstaller",
                                       "Workbench successfully installed. Please restart "
                                       "FreeCAD to apply the changes.")
            # symlink any macro contained in the module to the macros folder
            macro_dir = FreeCAD.getUserMacroDir(True)
            if not os.path.exists(macro_dir):
                os.makedirs(macro_dir)
            if os.path.exists(clonedir):
                for f in os.listdir(clonedir):
                    if f.lower().endswith(".fcmacro"):
                        print("copying macro:", f)
                        utils.symlink(os.path.join(clonedir, f), os.path.join(macro_dir, f))
                        FreeCAD.ParamGet('User parameter:Plugins/' +
                                         self.repos[idx][0]).SetString("destination", clonedir)
                        answer += "\n\n" + translate("AddonsInstaller",
                                                     "A macro has been installed and is available "
                                                     "under Macro -> Macros menu")
                        answer += ":\n<b>" + f + "</b>"
            self.progressbar_show.emit(False)
            self.info_label.emit(answer)
            self.mark_recompute.emit(self.repos[idx][0])
        self.stop = True

    def check_dependencies(self, baseurl):
        "checks if the repo contains a metadata.txt and check its contents"

        ok = True
        message = ""
        depsurl = baseurl.replace("github.com", "raw.githubusercontent.com")
        if not depsurl.endswith("/"):
            depsurl += "/"
        depsurl += "master/metadata.txt"
        mu = utils.urlopen(depsurl)
        if mu:
            # metadata.txt found
            depsfile = mu.read()
            mu.close()

            # urllib2 gives us a bytelike object instead of a string. Have to consider that
            try:
                depsfile = depsfile.decode("utf-8")
            except AttributeError:
                pass

            deps = depsfile.split("\n")
            for line in deps:
                if line.startswith("workbenches="):
                    depswb = line.split("=")[1].split(",")
                    for wb in depswb:
                        if wb.strip():
                            if not wb.strip() in FreeCADGui.listWorkbenches().keys():
                                if not wb.strip()+"Workbench" in FreeCADGui.listWorkbenches().keys():
                                    ok = False
                                    message += translate("AddonsInstaller", "Missing workbench") + ": " + wb + ", "
                elif line.startswith("pylibs="):
                    depspy = line.split("=")[1].split(",")
                    for pl in depspy:
                        if pl.strip():
                            try:
                                __import__(pl.strip())
                            except ImportError:
                                ok = False
                                message += translate("AddonsInstaller", "Missing python module") + ": " + pl + ", "
                elif line.startswith("optionalpylibs="):
                    opspy = line.split("=")[1].split(",")
                    for pl in opspy:
                        if pl.strip():
                            try:
                                __import__(pl.strip())
                            except ImportError:
                                message += translate("AddonsInstaller",
                                                     "Missing optional python module (doesn't prevent installing)")
                                message += ": " + pl + ", "
        if message and (not ok):
            message = translate("AddonsInstaller", "Some errors were found that prevent to install this workbench")
            message += ": <b>" + message + "</b>. "
            message += translate("AddonsInstaller", "Please install the missing components first.")
        return ok, message

    def download(self, baseurl, clonedir):
        "downloads and unzip a zip version from a git repo"

        bakdir = None
        if os.path.exists(clonedir):
            bakdir = clonedir+".bak"
            if os.path.exists(bakdir):
                shutil.rmtree(bakdir)
            os.rename(clonedir, bakdir)
        os.makedirs(clonedir)
        if baseurl in url_to_branch_map:
            branch = url_to_branch_map[baseurl]
        else:
            branch = "master"
        zipurl = utils.get_zip_url(baseurl, branch)
        if not zipurl:
            return translate("AddonsInstaller", "Error: Unable to locate zip from") + " " + baseurl
        try:
            print("Downloading "+zipurl)
            u = utils.urlopen(zipurl)
        except Exception:
            return translate("AddonsInstaller", "Error: Unable to download") + " " + zipurl
        if not u:
            return translate("AddonsInstaller", "Error: Unable to download") + " " + zipurl
        zfile = _stringio()
        zfile.write(u.read())
        zfile = zipfile.ZipFile(zfile)
        master = zfile.namelist()[0]  # github will put everything in a subfolder
        zfile.extractall(clonedir)
        u.close()
        zfile.close()
        for filename in os.listdir(clonedir+os.sep+master):
            shutil.move(clonedir+os.sep+master+os.sep+filename, clonedir+os.sep+filename)
        os.rmdir(clonedir+os.sep+master)
        if bakdir:
            shutil.rmtree(bakdir)
        return translate("AddonsInstaller", "Successfully installed") + " " + zipurl


class CheckSingleWorker(QtCore.QThread):
    """Worker to check for updates for a single addon"""

    updateAvailable = QtCore.Signal(bool)

    def __init__(self, name):

        QtCore.QThread.__init__(self)
        self.name = name

    def run(self):

        if not have_git:
            return
        FreeCAD.Console.PrintLog("Checking for available updates of the "+self.name+" addon\n")
        addondir = os.path.join(FreeCAD.getUserAppDataDir(), "Mod", self.name)
        if os.path.exists(addondir):
            if os.path.exists(addondir + os.sep + ".git"):
                gitrepo = git.Git(addondir)
                try:
                    gitrepo.fetch()
                    if "git pull" in gitrepo.status():
                        self.updateAvailable.emit(True)
                        return
                except Exception:
                    # can fail for any number of reasons, ex. not being online
                    pass
        self.updateAvailable.emit(False)

#  @}



# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create Label objects.

Labels are similar to text annotations but include a leader line
and an arrow in order to point to an object and indicate some of its
properties.
"""
## @package gui_labels
# \ingroup draftguitools
# \brief Provides GUI tools to create Label objects.

## \addtogroup draftguitools
# @{
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import DraftVecUtils
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
import draftguitools.gui_trackers as trackers
import draftutils.utils as utils

from draftutils.messages import _msg
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Label(gui_base_original.Creator):
    """Gui Command for the Label tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {'Pixmap': 'Draft_Label',
                'Accel': "D, L",
                'MenuText': QT_TRANSLATE_NOOP("Draft_Label", "Label"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Label", "Creates a label, optionally attached to a selected object or subelement.\n\nFirst select a vertex, an edge, or a face of an object, then call this command,\nand then set the position of the leader line and the textual label.\nThe label will be able to display information about this object, and about the selected subelement,\nif any.\n\nIf many objects or many subelements are selected, only the first one in each case\nwill be used to provide information to the label.")}

    def Activated(self):
        """Execute when the command is called."""
        self.name = translate("draft", "Label")
        super(Label, self).Activated(self.name, noplanesetup=True)
        self.ghost = None
        self.labeltype = utils.getParam("labeltype", "Custom")
        self.sel = Gui.Selection.getSelectionEx()
        if self.sel:
            self.sel = self.sel[0]
        self.ui.labelUi(self.name, callback=self.setmode)
        self.ui.xValue.setFocus()
        self.ui.xValue.selectAll()
        self.ghost = trackers.lineTracker()
        self.call = self.view.addEventCallback("SoEvent", self.action)
        _msg(translate("draft", "Pick target point"))
        self.ui.isCopy.hide()

    def setmode(self, i):
        """Set the type of label, if it is associated to an object."""
        self.labeltype = ["Custom", "Name", "Label", "Position",
                          "Length", "Area", "Volume", "Tag", "Material"][i]
        utils.setParam("labeltype", self.labeltype)

    def finish(self, closed=False, cont=False):
        """Finish the command."""
        if self.ghost:
            self.ghost.finalize()
        super(Label, self).finish()

    def create(self):
        """Create the actual object."""
        if len(self.node) == 3:
            targetpoint = self.node[0]
            basepoint = self.node[2]
            v = self.node[2].sub(self.node[1])
            dist = v.Length
            if hasattr(App, "DraftWorkingPlane"):
                h = App.DraftWorkingPlane.u
                n = App.DraftWorkingPlane.axis
                r = App.DraftWorkingPlane.getRotation().Rotation
            else:
                h = App.Vector(1, 0, 0)
                n = App.Vector(0, 0, 1)
                r = App.Rotation()

            if abs(DraftVecUtils.angle(v, h, n)) <= math.pi/4:
                direction = "Horizontal"
                dist = -dist
            elif abs(DraftVecUtils.angle(v, h, n)) >= math.pi*3/4:
                direction = "Horizontal"
            elif DraftVecUtils.angle(v, h, n) > 0:
                direction = "Vertical"
            else:
                direction = "Vertical"
                dist = -dist

            tp = DraftVecUtils.toString(targetpoint)
            sel = None
            if self.sel:
                sel = "FreeCAD.ActiveDocument." + self.sel.Object.Name

                if self.sel.SubElementNames:
                    sub = "'" + self.sel.SubElementNames[0] + "'"
                else:
                    sub = "None"

            pl = "FreeCAD.Placement"
            pl += "("
            pl += DraftVecUtils.toString(basepoint) + ", "
            pl += "FreeCAD.Rotation" + str(r.Q)
            pl += ")"

            Gui.addModule("Draft")
            _cmd = "Draft.make_label"
            _cmd += "("
            _cmd += "target_point=" + tp + ", "
            _cmd += "placement=" + pl + ", "
            if sel:
                _cmd += "target_object=" + sel + ", "
                _cmd += "subelements=" + sub + ", "
            _cmd += "label_type=" + "'" + self.labeltype + "'" + ", "
            # _cmd += "custom_text=" + "'Label'" + ", "
            _cmd += "direction=" + "'" + direction + "'" + ", "
            _cmd += "distance=" + str(dist)
            _cmd += ")"

            # Commit the creation instructions through the parent class,
            # the Creator class
            _cmd_list = ['_label_ = ' + _cmd,
                         'Draft.autogroup(_label_)',
                         'FreeCAD.ActiveDocument.recompute()']
            self.commit(translate("draft", "Create Label"),
                        _cmd_list)
        self.finish()

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif arg["Type"] == "SoLocation2Event":
            if hasattr(Gui, "Snapper"):
                Gui.Snapper.affinity = None  # don't keep affinity
            if len(self.node) == 2:
                gui_tool_utils.setMod(arg, gui_tool_utils.MODCONSTRAIN, True)
            self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg)
            gui_tool_utils.redraw3DView()
        elif arg["Type"] == "SoMouseButtonEvent":
            if (arg["State"] == "DOWN") and (arg["Button"] == "BUTTON1"):
                if self.point:
                    self.ui.redraw()
                    if not self.node:
                        # first click
                        self.node.append(self.point)
                        self.ui.isRelative.show()
                        _msg(translate("draft",
                                       "Pick endpoint of leader line"))
                        if self.planetrack:
                            self.planetrack.set(self.point)
                    elif len(self.node) == 1:
                        # second click
                        self.node.append(self.point)
                        if self.ghost:
                            self.ghost.p1(self.node[0])
                            self.ghost.p2(self.node[1])
                            self.ghost.on()
                        _msg(translate("draft", "Pick text position"))
                    else:
                        # third click
                        self.node.append(self.point)
                        self.create()

    def numericInput(self, numx, numy, numz):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.point = App.Vector(numx, numy, numz)
        if not self.node:
            # first click
            self.node.append(self.point)
            self.ui.isRelative.show()
            _msg(translate("draft", "Pick endpoint of leader line"))
            if self.planetrack:
                self.planetrack.set(self.point)
        elif len(self.node) == 1:
            # second click
            self.node.append(self.point)
            if self.ghost:
                self.ghost.p1(self.node[0])
                self.ghost.p2(self.node[1])
                self.ghost.on()
            _msg(translate("draft", "Pick text position"))
        else:
            # third click
            self.node.append(self.point)
            self.create()


Draft_Label = Label
Gui.addCommand('Draft_Label', Label())

## @}



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Elmer equation object Flux"
__author__ = "Markus Hovorka"
__url__ = "https://www.freecadweb.org"

## \addtogroup FEM
#  @{

from femtools import femutils
from ... import equationbase
from . import linear


def create(doc, name="Flux"):
    return femutils.createObject(
        doc, name, Proxy, ViewProxy)


class Proxy(linear.Proxy, equationbase.FluxProxy):

    Type = "Fem::EquationElmerFlux"

    def __init__(self, obj):
        super(Proxy, self).__init__(obj)
        obj.addProperty(
            "App::PropertyBool",
            "CalculateFlux",
            "Flux",
            ""
        )
        obj.addProperty(
            "App::PropertyString",
            "FluxVariable",
            "Flux",
            "Insert variable name for flux calculation"
        )
        """
        obj.addProperty(
            "App::PropertyBool",
            "CalculateFluxAbs",
            "Flux",
            "Select calculation of abs of flux"
        )
        obj.addProperty(
            "App::PropertyBool",
            "CalculateFluxMagnitude",
            "Flux",
            "Select calculation of magnitude of flux"
        )
        """
        obj.addProperty(
            "App::PropertyBool",
            "CalculateGrad",
            "Flux",
            "Select calculation of gradient"
        )
        """
        obj.addProperty(
            "App::PropertyBool",
            "CalculateGradAbs",
            "Flux",
            "Select calculation of abs of gradient"
        )
        obj.addProperty(
            "App::PropertyBool",
            "CalculateGradMagnitude",
            "Flux",
            "Select calculation of magnitude of gradient"
        )
        obj.addProperty(
            "App::PropertyBool",
            "EnforcePositiveMagnitude",
            "Flux",
            "Select calculation of positive magnitude"
        )
        """
        obj.Priority = 5


class ViewProxy(linear.ViewProxy, equationbase.FluxViewProxy):
    pass

##  @}
